
/*****************************************************************************/
/*                                                                           */
/*  THE NRCONV NURSE ROSTERING TO XESTT CONVERTER                            */
/*  COPYRIGHT (C) 2016, Jeffrey H. Kingston                                  */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         nrc_instance_metadata.c                                    */
/*  DESCRIPTION:  Instance metadata                                          */
/*                                                                           */
/*****************************************************************************/
#include <stdarg.h>
#include "nrc_interns.h"


/*****************************************************************************/
/*                                                                           */
/*  KHE_INSTANCE_METADATA - instance metadata                                */
/*                                                                           */
/*****************************************************************************/

struct nrc_instance_metadata_rec {
  char			*name;			/* instance name             */
  char			*contributor;		/* contributor               */
  char			*date;			/* date                      */
  char			*country;		/* country                   */
  char			*description;		/* description               */
  char			*remarks;		/* remarks (optional)        */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "construction and query"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  NRC_INSTANCE_METADATA NrcInstanceMetaDataMake(char *name, char *contr,   */
/*    char *date, char *country, char *description, char *remarks)           */
/*                                                                           */
/*  Make instance metadata with these attributes.                            */
/*                                                                           */
/*****************************************************************************/

NRC_INSTANCE_METADATA NrcInstanceMetaDataMake(char *name, char *contributor,
  char *date, char *country, char *description, char *remarks)
{
  NRC_INSTANCE_METADATA res;
  MMake(res);
  res->name = name;
  res->contributor = contributor;
  res->date = date;
  res->country = country;
  res->description = description;
  res->remarks = remarks;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcInstanceMetaDataName(NRC_INSTANCE_METADATA md)                  */
/*                                                                           */
/*  Return the Name attribute of md.                                         */
/*                                                                           */
/*****************************************************************************/

char *NrcInstanceMetaDataName(NRC_INSTANCE_METADATA md)
{
  return md->name;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcInstanceMetaDataContributor(NRC_INSTANCE_METADATA md)           */
/*                                                                           */
/*  Return the Contributor attribute of md.                                  */
/*                                                                           */
/*****************************************************************************/

char *NrcInstanceMetaDataContributor(NRC_INSTANCE_METADATA md)
{
  return md->contributor;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcInstanceMetaDataDate(NRC_INSTANCE_METADATA md)                  */
/*                                                                           */
/*  Return the Date attribute of md.                                         */
/*                                                                           */
/*****************************************************************************/

char *NrcInstanceMetaDataDate(NRC_INSTANCE_METADATA md)
{
  return md->date;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcInstanceMetaDataCountry(NRC_INSTANCE_METADATA md)               */
/*                                                                           */
/*  Return the Country attribute of md.                                      */
/*                                                                           */
/*****************************************************************************/

char *NrcInstanceMetaDataCountry(NRC_INSTANCE_METADATA md)
{
  return md->country;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcInstanceMetaDataDescription(NRC_INSTANCE_METADATA md)           */
/*                                                                           */
/*  Return the Description attribute of md.                                  */
/*                                                                           */
/*****************************************************************************/

char *NrcInstanceMetaDataDescription(NRC_INSTANCE_METADATA md)
{
  return md->description;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcInstanceMetaDataRemarks(NRC_INSTANCE_METADATA md)               */
/*                                                                           */
/*  Return the optional Remarks attribute of md.                             */
/*                                                                           */
/*****************************************************************************/

char *NrcInstanceMetaDataRemarks(NRC_INSTANCE_METADATA md)
{
  return md->remarks;
}
