
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_group_by_frame.c                                    */
/*  DESCRIPTION:  Group by frame                                             */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include <limits.h>
#define bool_show(x) ((x) ? "true" : "false")
#define min(a, b) ((a) < (b) ? (a) : (b))

#define DEBUG1 0
#define DEBUG2 0  /* KheDoCombinatorialGroupingForInterval */
#define DEBUG3 0  /* shows which tasks are being grouped */
#define DEBUG4 0
#define DEBUG5 0  /* KheTimeBasedGrouping (detailed) */
#define DEBUG6 0  /* phase 0 */
#define	DEBUG7 0  /* profile grouping */
#define DEBUG8 0  /* KheTimeBasedGrouping (brief) */
#define DEBUG9 0  /* KheDoProfileGroupingMonitors */
#define DEBUG10 0  /* KheDoProfile */
#define DEBUG11 0


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_GROUP_INFO                                                      */
/*                                                                           */
/*  Everything about what is going on at one time group (one day).           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_time_group_info_rec {
  KHE_TIME_GROUP		time_group;		/* the time group    */
  bool				linked_to_prev;		/* linked to prev tg */
} *KHE_TIME_GROUP_INFO;

typedef HA_ARRAY(KHE_TIME_GROUP_INFO) ARRAY_KHE_TIME_GROUP_INFO;


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_INFO, KHE_ROOT_TASK_SET, KHE_RCG_SOLVER                         */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_root_task_set_rec *KHE_ROOT_TASK_SET;
typedef HA_ARRAY(KHE_ROOT_TASK_SET) ARRAY_KHE_ROOT_TASK_SET;

typedef struct khe_time_info_rec *KHE_TIME_INFO; 
typedef HA_ARRAY(KHE_TIME_INFO) ARRAY_KHE_TIME_INFO;

typedef struct khe_rcg_solver_rec *KHE_RCG_SOLVER;


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_INFO                                                            */
/*                                                                           */
/*  Everything about what is going on at one time.                           */
/*                                                                           */
/*****************************************************************************/

struct khe_time_info_rec {
  KHE_RCG_SOLVER	solver;			/* enclosing solver         */
  KHE_TIME		time;			/* the time                 */
  ARRAY_KHE_ROOT_TASK_SET root_task_sets;	/* root tasks at time       */
  bool			pg_active;		/* pg active now            */
  int			pg_profile;		/* profile at this time     */
};


/*****************************************************************************/
/*                                                                           */
/*  KHE_ROOT_TASK_SET                                                        */
/*                                                                           */
/*  A set of equivalent root tasks.                                          */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_PG_INACTIVE_UNMONITORED,	/* inactive because unmonitored              */
  KHE_PG_INACTIVE_MAXIMAL,	/* inactive because maximal intersection     */
  KHE_PG_ACTIVE			/* active (contributes to profile)           */
} KHE_PG_STATUS;

struct khe_root_task_set_rec {
  KHE_RCG_SOLVER	solver;			/* enclosing solver         */
  KHE_TASK_SET		tasks;			/* the tasks                */
  KHE_RESOURCE_GROUP	domain;			/* their common domain      */
  KHE_RESOURCE		init_asst;		/* if assigned initially    */
  KHE_TIME_SET		time_set;		/* their common times       */
  ARRAY_KHE_TIME_INFO	time_info;		/* parallels time_set       */
  bool			ignore_for_leader;	/* if ignoring for root_rts */
  KHE_PG_STATUS		pg_status;		/* contributes to profile   */
};


/*****************************************************************************/
/*                                                                           */
/*  KHE_RCG_SOLVER - solver for group by resource constraint solving         */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_TIME_SET) ARRAY_KHE_TIME_SET;
typedef HA_ARRAY(KHE_TASK_SET) ARRAY_KHE_TASK_SET;
typedef HA_ARRAY(KHE_LIMIT_ACTIVE_INTERVALS_MONITOR)
  ARRAY_KHE_LIMIT_ACTIVE_INTERVALS_MONITOR;
typedef HA_ARRAY(KHE_RESOURCE_GROUP) ARRAY_KHE_RESOURCE_GROUP;

struct khe_rcg_solver_rec {

  /* the solve as a whole */
  HA_ARENA			arena;
  KHE_SOLN			soln;
  KHE_INSTANCE			instance;
  bool				resource_invariant;
  KHE_FRAME			common_frame;
  KHE_EVENT_TIMETABLE_MONITOR	etm;
  KHE_RESOURCE_TYPE		resource_type;
  /* KHE_RESOURCE		resource; */
  bool				soln_changed;
  KHE_TASK_SET			task_set;

  /* what's happening at each time group, and at each time */
  ARRAY_KHE_TIME_GROUP_INFO	time_group_info;
  ARRAY_KHE_TIME_INFO		time_info;
  ARRAY_KHE_ROOT_TASK_SET	free_root_task_sets;

  /* time-based grouping */
  ARRAY_KHE_ROOT_TASK_SET	tbg_rts_set;
  ARRAY_KHE_ROOT_TASK_SET	tbg_ignore_for_leader_set;
  KHE_TIME_SET			tbg_in_ts;
  KHE_TIME_SET			tbg_out_ts;

  /* framed monitors */
  /* ***
  ARRAY_KHE_FRAMED_MONITOR	framed_monitors;
  HA_ARRAY_INT			start_indexes;
  *** */

  /* eliminating combinations */
  ARRAY_KHE_TIME_GROUP_INFO	ec_time_group_info;
  HA_ARRAY_INT			ec_intersects;

  /* combinatorial grouping */
  int				cg_max_days;
  /* KHE_GROUP_MONITOR		cg_group_monitor; */
  int				cg_first_index;
  int				cg_last_index;
  KHE_TIME_SET			cg_in_ts;
  KHE_TIME_SET			cg_out_ts;
  int				cg_success_count;
  KHE_TIME_SET			cg_success_in_ts;
  KHE_TIME_SET			cg_success_out_ts;
  KHE_TRACE			cg_trace;

  /* profile grouping */
  bool				pg_active;
  ARRAY_KHE_LIMIT_ACTIVE_INTERVALS_MONITOR pg_monitors;
  KHE_TIME_SET			pg_monitor_ts;
  int				pg_min_limit;
  int				pg_max_limit;
  bool				pg_history_before;
  bool				pg_history_after;
  KHE_TIME_SET			pg_in_ts;
  KHE_TIME_SET			pg_out_ts;
  ARRAY_KHE_RESOURCE_GROUP	pg_domains;
  HA_ARRAY_INT			pg_domain_counts;

  /* where does this belong? */
  /* ARRAY_KHE_TASK_SET		task_sets; */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "time group info objects"                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_GROUP_INFO KheTimeGroupInfoMake(KHE_TIME_GROUP tg, HA_ARENA a)  */
/*                                                                           */
/*  Make a new time group info object for tg.                                */
/*                                                                           */
/*****************************************************************************/

static KHE_TIME_GROUP_INFO KheTimeGroupInfoMake(KHE_TIME_GROUP tg, HA_ARENA a)
{
  KHE_TIME_GROUP_INFO res;
  HaMake(res, a);
  res->time_group = tg;
  res->linked_to_prev = false;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeGroupInfoDebug(KHE_TIME_GROUP_INFO tgi, int indent, FILE *fp)*/
/*                                                                           */
/*  Debug print of tgi with the given indent.                                */
/*                                                                           */
/*****************************************************************************/

static void KheTimeGroupInfoDebug(KHE_TIME_GROUP_INFO tgi, int indent, FILE *fp)
{
  fprintf(fp, "%*s[ TimeGroupInfo %s", indent, "",
    tgi->linked_to_prev ? "linked " : "");
  KheTimeGroupDebug(tgi->time_group, 1, 0, fp);
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "time info objects"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_INFO KheTimeInfoMake(KHE_RCG_SOLVER gs, KHE_TIME t)             */
/*                                                                           */
/*  Make a new time info object for t.                                       */
/*                                                                           */
/*****************************************************************************/

static KHE_TIME_INFO KheTimeInfoMake(KHE_RCG_SOLVER gs, KHE_TIME t)
{
  KHE_TIME_INFO res;
  HaMake(res, gs->arena);
  res->solver = gs;
  res->time = t;
  HaArrayInit(res->root_task_sets, gs->arena);
  res->pg_active = false;
  res->pg_profile = -1;  /* undefined unless pg_active */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeInfoAddRootTaskSet(KHE_TIME_INFO ti, KHE_ROOT_TASK_SET rts)  */
/*                                                                           */
/*  Add rts to ti, at a suitable spot for its domain and assigned resource.  */
/*                                                                           */
/*  Implementation note.  This function is called only when rts has no       */
/*  tasks, and so rts has no implications for time info profiles.            */
/*                                                                           */
/*****************************************************************************/
static bool KheRootTaskSetLT(KHE_ROOT_TASK_SET rts1, KHE_ROOT_TASK_SET rts2);

static void KheTimeInfoAddRootTaskSet(KHE_TIME_INFO ti, KHE_ROOT_TASK_SET rts)
{
  int i;  KHE_ROOT_TASK_SET rts2;
  for( i = 0;  i < HaArrayCount(ti->root_task_sets);  i++ )
  {
    rts2 = HaArray(ti->root_task_sets, i);
    if( KheRootTaskSetLT(rts, rts2) )
      break;
  }
  HaArrayAdd(ti->root_task_sets, i, rts);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeInfoDebug(KHE_TIME_INFO ti, int indent, FILE *fp)            */
/*                                                                           */
/*  Debug print of ti with the given indent.                                 */
/*                                                                           */
/*****************************************************************************/
static void KheRootTaskSetDebug(KHE_ROOT_TASK_SET rts, int indent, FILE *fp);

static void KheTimeInfoDebug(KHE_TIME_INFO ti, int indent, FILE *fp)
{
  KHE_ROOT_TASK_SET rts;  int i;
  fprintf(fp, "%*s[ TimeInfo(%s", indent, "", KheTimeId(ti->time));
  if( ti->pg_active )
    fprintf(fp, ", pg_profile %d", ti->pg_profile);
  fprintf(fp, ")\n");
  HaArrayForEach(ti->root_task_sets, rts, i)
    KheRootTaskSetDebug(rts, indent + 2, fp);
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "root task sets"                                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_PG_STATUS KheRootTaskSetStatus(KHE_ROOT_TASK_SET rts)                */
/*                                                                           */
/*  Return the profile grouping status of rts.                               */
/*                                                                           */
/*  Implementation note.  When this function is called when gs->pg_active,   */
/*  it is only called on root task sets with a non-empty intersection with   */
/*  gs->pg_monitor_ts, hence the assertion to this effect.                   */
/*                                                                           */
/*****************************************************************************/

KHE_PG_STATUS KheRootTaskSetStatus(KHE_ROOT_TASK_SET rts)
{
  KHE_RCG_SOLVER gs;  int count;
  gs = rts->solver;
  if( gs->pg_active )
  {
    count = KheTimeSetIntersectCount(gs->pg_monitor_ts, rts->time_set);
    HnAssert(count > 0, "KheRootTaskSetStatus internal error");
    if( count >= gs->pg_max_limit )
      return KHE_PG_INACTIVE_MAXIMAL;
    else
      return KHE_PG_ACTIVE;
  }
  else
    return KHE_PG_INACTIVE_UNMONITORED;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_ROOT_TASK_SET KheRootTaskSetMake(KHE_RCG_SOLVER gs, KHE_TASK task)   */
/*                                                                           */
/*  Make a new root task set suitable for holding task (but do not add task  */
/*  to it), and add it to the appropriate time info objects.                 */
/*                                                                           */
/*  Implementation note.  Because the new object contains no tasks when      */
/*  it returns, its presence has no implications for time info profiles.     */
/*                                                                           */
/*****************************************************************************/

static KHE_ROOT_TASK_SET KheRootTaskSetMake(KHE_RCG_SOLVER gs, KHE_TASK task)
{
  KHE_ROOT_TASK_SET res;  int i;  KHE_TIME t;  KHE_TIME_INFO ti;  
  if( HaArrayCount(gs->free_root_task_sets) > 0 )
  {
    res = HaArrayLastAndDelete(gs->free_root_task_sets);
    KheTaskSetClear(res->tasks);
    KheTimeSetClear(res->time_set);
    HaArrayClear(res->time_info);
  }
  else
  {
    HaMake(res, gs->arena);
    res->tasks = KheTaskSetMake(gs->soln);
    res->time_set = KheTimeSetMake(KheSolnInstance(gs->soln), gs->arena);
    HaArrayInit(res->time_info, gs->arena);
  }
  res->solver = gs;
  res->domain = KheTaskDomain(task);
  res->init_asst = KheTaskAsstResource(task);
  KheTimeSetAddTaskTimes(res->time_set, task);
  for( i = 0;  i < KheTimeSetTimeCount(res->time_set);  i++ )
  {
    t = KheTimeSetTime(res->time_set, i);
    ti = HaArray(gs->time_info, KheTimeIndex(t));
    HaArrayAddLast(res->time_info, ti);
    KheTimeInfoAddRootTaskSet(ti, res);
  }
  res->ignore_for_leader = false;
  res->pg_status = KheRootTaskSetStatus(res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRootTaskSetDelete(KHE_ROOT_TASK_SET rts)                         */
/*                                                                           */
/*  Delete rts from its enclosing solver.                                    */
/*                                                                           */
/*  Implementation note.  Here rts must be empty before it is deleted,       */
/*  and so the deletion has no consequences for profile grouping.            */
/*                                                                           */
/*****************************************************************************/

static void KheRootTaskSetDelete(KHE_ROOT_TASK_SET rts)
{
  int i, j;  KHE_TIME_INFO ti;  KHE_ROOT_TASK_SET rts2;
  HnAssert(KheTaskSetTaskCount(rts->tasks) == 0,
    "KheRootTaskSetDelete internal error");
  HaArrayForEach(rts->time_info, ti, i)
  {
    HaArrayForEach(ti->root_task_sets, rts2, j)
      if( rts2 == rts )
	break;
    HnAssert(j < HaArrayCount(ti->root_task_sets),
      "KheRootTaskSetDelete internal error 1");
    HaArrayDeleteAndShift(ti->root_task_sets, j);
  }
  HaArrayAddLast(rts->solver->free_root_task_sets, rts);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheRootTaskSetLT(KHE_ROOT_TASK_SET rts1, KHE_ROOT_TASK_SET rts2)    */
/*                                                                           */
/*  Return true if rts1 comes earlier in sequence than rts2.                 */
/*                                                                           */
/*****************************************************************************/

static bool KheRootTaskSetLT(KHE_ROOT_TASK_SET rts1, KHE_ROOT_TASK_SET rts2)
{
  int rts1_val, rts2_val;
  rts1_val = (rts1->init_asst != NULL ? 0 : 1);
  rts2_val = (rts2->init_asst != NULL ? 0 : 1);
  return rts1_val != rts2_val ? rts1_val < rts2_val :
    KheResourceGroupResourceCount(rts1->domain) <
    KheResourceGroupResourceCount(rts2->domain);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheRootTaskSetAcceptsTask(KHE_ROOT_TASK_SET rts, KHE_TASK task)     */
/*                                                                           */
/*  Return true if task is equivalent to the tasks already in rts, and       */
/*  so could be added to it.                                                 */
/*                                                                           */
/*****************************************************************************/

static bool KheRootTaskSetAcceptsTask(KHE_ROOT_TASK_SET rts, KHE_TASK task)
{
  return KheTaskAsstResource(task) == rts->init_asst &&
    KheTaskEquivalent(KheTaskSetTask(rts->tasks, 0), task);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRootTaskSetAddTask(KHE_ROOT_TASK_SET rts, KHE_TASK task)         */
/*                                                                           */
/*  Add task to rts, assuming KheRootTaskSetAcceptsTask(rts, task) holds.    */
/*                                                                           */
/*****************************************************************************/

static void KheRootTaskSetAddTask(KHE_ROOT_TASK_SET rts, KHE_TASK task)
{
  KHE_TIME_INFO ti;  int i;
  HnAssert(KheTaskAsst(task)==NULL, "KheRootTaskSetAddTask: task is assigned");
  KheTaskSetAddTask(rts->tasks, task);
  if( rts->pg_status == KHE_PG_ACTIVE )
    HaArrayForEach(rts->time_info, ti, i)
      if( ti->pg_active )
        ti->pg_profile++;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_TASK KheRootTaskSetLastAndDelete(KHE_ROOT_TASK_SET rts)              */
/*                                                                           */
/*  Delete and return the last task of rts.  If this causes rts to become    */
/*  empty, then delete rts from the enclosing solver.                        */
/*                                                                           */
/*****************************************************************************/

static KHE_TASK KheRootTaskSetLastAndDelete(KHE_ROOT_TASK_SET rts)
{
  int i;  KHE_TIME_INFO ti;  KHE_TASK res;
  res = KheTaskSetLastAndDelete(rts->tasks);
  if( rts->pg_status == KHE_PG_ACTIVE )
    HaArrayForEach(rts->time_info, ti, i)
      if( ti->pg_active )
	ti->pg_profile--;
  if( KheTaskSetTaskCount(rts->tasks) == 0 )
    KheRootTaskSetDelete(rts);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRootTaskSetDebug(KHE_ROOT_TASK_SET rts, int indent, FILE *fp)    */
/*                                                                           */
/*  Debug print of rts with the given indent.                                */
/*                                                                           */
/*****************************************************************************/

static void KheRootTaskSetDebug(KHE_ROOT_TASK_SET rts, int indent, FILE *fp)
{
  fprintf(fp, "%*s[ RootTaskSet(%s", indent, "",
    rts->pg_status == KHE_PG_INACTIVE_UNMONITORED ? "" :
    rts->pg_status == KHE_PG_INACTIVE_MAXIMAL ? "maximal, " : "active, ");
  KheResourceGroupDebug(rts->domain, 1, -1, fp);
  if( rts->init_asst != NULL )
    fprintf(fp, " := %s", KheResourceId(rts->init_asst));
  fprintf(fp, ", ");
  KheTimeSetDebug(rts->time_set, 1, -1, fp);
  fprintf(fp, "%d tasks): ", KheTaskSetTaskCount(rts->tasks));
  KheTaskSetDebug(rts->tasks, 2, -1, fp);
  fprintf(fp, " ]\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "group by resource constraints solvers"                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheGroupSolverAddTask(KHE_RCG_SOLVER gs, KHE_TASK task)             */
/*                                                                           */
/*  Add task to gs and return true, or if task has no assigned time, do      */
/*  nothing.  Here task must be a proper root.                               */
/*                                                                           */
/*****************************************************************************/

static bool KheGroupSolverAddTask(KHE_RCG_SOLVER gs, KHE_TASK task)
{
  KHE_MEET meet;  KHE_TIME t;  int i;  KHE_TIME_INFO ti;  KHE_ROOT_TASK_SET rts;
  HnAssert(KheTaskAsst(task) == NULL, "KheGroupSolverAddTask internal error");
  meet = KheTaskMeet(task);
  if( meet != NULL )
  {
    t = KheMeetAsstTime(meet);
    if( t != NULL )
    {
      /* search for a set of equivalent root tasks */
      ti = HaArray(gs->time_info, KheTimeIndex(t));
      rts = NULL;  /* keep compiler happy */
      HaArrayForEach(ti->root_task_sets, rts, i)
	if( KheRootTaskSetAcceptsTask(rts, task) )
	  break;

      /* if no equivalent tasks, make a new set (initially empty) */
      if( i >= HaArrayCount(ti->root_task_sets) )
	rts = KheRootTaskSetMake(gs, task);

      /* add task to the set (old or new) and return true */
      KheRootTaskSetAddTask(rts, task);
      return true;
    }
  }
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_RCG_SOLVER KheGroupByResourceConstraintsSolverMake(                  */
/*    KHE_SOLN soln, KHE_RESOURCE_TYPE rt, int max_days,                     */
/*    bool resource_invariant, KHE_FRAME common_frame,                       */
/*    KHE_EVENT_TIMETABLE_MONITOR etm, KHE_TASK_SET ts, HA_ARENA a)          */
/*                                                                           */
/*  Make a new group by resource constraints solver object with these        */
/*  attributes.  Add all tasks which are either unassigned or assigned       */
/*  directly to cycle tasks (that is, tasks which are proper roots).         */
/*                                                                           */
/*****************************************************************************/

static KHE_RCG_SOLVER KheGroupByResourceConstraintsSolverMake(
  KHE_SOLN soln, KHE_RESOURCE_TYPE rt, int max_days,
  bool resource_invariant, KHE_FRAME common_frame,
  KHE_EVENT_TIMETABLE_MONITOR etm, KHE_TASK_SET ts, HA_ARENA a)
{
  KHE_RCG_SOLVER res;  int i;  KHE_TIME_GROUP tg;  KHE_TASK task;
  KHE_TIME_GROUP_INFO tgi;  KHE_TIME t;  KHE_TIME_INFO ti;

  /* the solve as a whole */
  HaMake(res, a);
  res->arena = a;
  res->soln = soln;
  res->instance = KheSolnInstance(soln);
  res->resource_invariant = resource_invariant;
  res->common_frame = common_frame;
  res->etm = etm;
  res->resource_type = rt;
  /* res->resource = NULL; */
  res->soln_changed = false;
  res->task_set = ts;

  /* time group info */
  HaArrayInit(res->time_group_info, a);
  for( i = 0;  i < KheFrameTimeGroupCount(common_frame);  i++ )
  {
    tg = KheFrameTimeGroup(common_frame, i);
    tgi = KheTimeGroupInfoMake(tg, a);
    HaArrayAddLast(res->time_group_info, tgi);
  }

  /* time info */
  HaArrayInit(res->time_info, a);
  for( i = 0;  i < KheInstanceTimeCount(res->instance);  i++ )
  {
    t = KheInstanceTime(res->instance, i);
    ti = KheTimeInfoMake(res, t);
    HaArrayAddLast(res->time_info, ti);
  }
  HaArrayInit(res->free_root_task_sets, a);

  /* time-based grouping */
  HaArrayInit(res->tbg_rts_set, a);
  HaArrayInit(res->tbg_ignore_for_leader_set, a);
  res->tbg_in_ts = KheTimeSetMake(res->instance, a);
  res->tbg_out_ts = KheTimeSetMake(res->instance, a);

  /* add unassigned tasks of the right type to time group info */
  for( i = 0;  i < KheSolnTaskCount(soln);  i++ )
  {
    task = KheSolnTask(soln, i);
    if( !KheTaskIsCycleTask(task) && KheTaskResourceType(task) == rt &&
	(KheTaskAsst(task) == NULL || KheTaskIsCycleTask(KheTaskAsst(task))) &&
	KheTaskNonAssignmentHasCost(task) )
      KheGroupSolverAddTask(res, task);
  }

  /* framed monitors */
  /* ***
  HaArrayInit(res->framed_monitors, a);
  HaArrayInit(res->start_indexes, a);
  *** */

  /* eliminating combinations */
  HaArrayInit(res->ec_time_group_info, a);
  HaArrayInit(res->ec_intersects, a);

  /* combinatorial grouping */
  res->cg_max_days = max_days;
  /* res->cg_group_monitor = NULL; */
  res->cg_first_index = -1;
  res->cg_last_index = -1;
  res->cg_in_ts = KheTimeSetMake(res->instance, a);
  res->cg_out_ts = KheTimeSetMake(res->instance, a);
  res->cg_success_count = 0;
  res->cg_success_in_ts = KheTimeSetMake(res->instance, a);
  res->cg_success_out_ts = KheTimeSetMake(res->instance, a);
  res->cg_trace = KheTraceMake((KHE_GROUP_MONITOR) soln);

  /* profile grouping */
  res->pg_active = false;
  HaArrayInit(res->pg_monitors, a);
  res->pg_monitor_ts = KheTimeSetMake(res->instance, a);
  res->pg_min_limit = -1;
  res->pg_max_limit = -1;
  res->pg_history_before = false;
  res->pg_history_after = false;
  res->pg_in_ts = KheTimeSetMake(res->instance, a);
  res->pg_out_ts = KheTimeSetMake(res->instance, a);
  HaArrayInit(res->pg_domains, a);
  HaArrayInit(res->pg_domain_counts, a);

  /* ? */
  /* HaArrayInit(res->task_sets, a); */

  /* all done */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByResourceConstraintsSolverDebug(KHE_RCG_SOLVER gs,         */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of group by resource constraints solver gs onto fp with      */
/*  the given indent.                                                        */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByResourceConstraintsSolverDebug(KHE_RCG_SOLVER gs,
  int indent, FILE *fp)
{
  int i;  KHE_TIME_GROUP_INFO tgi;  KHE_TIME_INFO ti;
  fprintf(fp, "%*s[ %s(%s, max_days %d, rs_invt %s)\n", 
    indent, "", "GroupByResourceConstraintsSolver", KheInstanceId(gs->instance),
    gs->cg_max_days, bool_show(gs->resource_invariant));
  fprintf(fp, "%*s[ time group info:\n", indent + 2, "");
  HaArrayForEach(gs->time_group_info, tgi, i)
    KheTimeGroupInfoDebug(tgi, indent + 4, fp);
  fprintf(fp, "%*s]\n", indent + 2, "");
  fprintf(fp, "%*s[ time info:\n", indent + 2, "");
  HaArrayForEach(gs->time_info, ti, i)
    KheTimeInfoDebug(ti, indent + 4, fp);
  fprintf(fp, "%*s]\n", indent + 2, "");
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "time-based grouping"                                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheRootTaskSetAcceptable(KHE_ROOT_TASK_SET rts,                     */
/*    bool check_ignore, KHE_TIME_SET nsubset_ts, KHE_TIME_SET disjoint_ts,  */
/*    KHE_ROOT_TASK_SET ass_to_rts)                                          */
/*                                                                           */
/*  Return true if rts passes the four tests.                                */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused
static bool KheRootTaskSetAcceptable(KHE_ROOT_TASK_SET rts,
  bool check_ignore, KHE_TIME_SET nsubset_ts, KHE_TIME_SET disjoint_ts,
  KHE_ROOT_TASK_SET ass_to_rts)
{
  return (!check_ignore || rts->ignore_for_leader == false) &&
    (nsubset_ts == NULL || !KheTimeSetSubset(nsubset_ts, rts->time_set)) &&
    (disjoint_ts == NULL || KheTimeSetDisjoint(disjoint_ts, rts->time_set)) &&
    (ass_to_rts == NULL || KheTaskAssignCheck(
      KheTaskSetTask(rts->tasks, 0), KheTaskSetTask(ass_to_rts->tasks, 0)));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheTimeFindMinDomainRootTaskSet(KHE_RCG_SOLVER gs,                  */
/*    KHE_TIME t, bool check_ignore, KHE_TIME_SET nsubset_ts,                */
/*    KHE_TIME_SET disjoint_ts, KHE_ROOT_TASK_SET ass_to_rts,                */
/*    KHE_ROOT_TASK_SET *res)                                                */
/*                                                                           */
/*  Like KheFindMinDomainRootTaskSet below, only for a single time, not      */
/*  for a whole time set.                                                    */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused
static bool KheTimeFindMinDomainRootTaskSet(KHE_RCG_SOLVER gs,
  KHE_TIME t, bool check_ignore, KHE_TIME_SET nsubset_ts,
  KHE_TIME_SET disjoint_ts, KHE_ROOT_TASK_SET ass_to_rts,
  KHE_ROOT_TASK_SET *res)
{
  KHE_ROOT_TASK_SET rts;  int i;  KHE_TIME_INFO ti;
  ti = HaArray(gs->time_info, KheTimeIndex(t));
  HaArrayForEach(ti->root_task_sets, rts, i)
    if( KheRootTaskSetAcceptable(rts, check_ignore, nsubset_ts, disjoint_ts,
	  ass_to_rts) )
      return *res = rts, true;
  return *res = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheFindMinDomainRootTaskSet(KHE_RCG_SOLVER gs,                      */
/*    KHE_TIME_SET ts, bool check_ignore, KHE_TIME_SET nsubset_ts,           */
/*    KHE_TIME_SET disjoint_ts, KHE_ROOT_TASK_SET ass_to_rts,                */
/*    KHE_ROOT_TASK_SET *res)                                                */
/*                                                                           */
/*  Find a root task set of minimum-sized domain in ts, subject to           */
/*  four conditions:                                                         */
/*                                                                           */
/*    * if check_ignore is true, the ignore_for_leader field for the set     */
/*      must be false;                                                       */
/*                                                                           */
/*    * if nsubset_ts != NULL, the times of nsubset_ts must not be a subset  */
/*      the times of *res;                                                   */
/*                                                                           */
/*    * if disjoint_ts != NULL, the times of disjoint_ts must be disjoint    */
/*      from the times of *res;                                              */
/*                                                                           */
/*    * if ass_to_rts, the tasks of *res must be assignable to the           */
/*      tasks of ass_to_rts.                                                 */
/*                                                                           */
/*  If successful, set *res to the root task set and return true, otherwise  */
/*  set *res to NULL and return false.                                       */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused
static bool KheFindMinDomainRootTaskSet(KHE_RCG_SOLVER gs,
  KHE_TIME_SET ts, bool check_ignore, KHE_TIME_SET nsubset_ts,
  KHE_TIME_SET disjoint_ts, KHE_ROOT_TASK_SET ass_to_rts,
  KHE_ROOT_TASK_SET *res)
{
  KHE_ROOT_TASK_SET rts;  int i, j, res_count, rts_count;
  KHE_TIME t;  KHE_TIME_INFO ti;
  *res = NULL;
  res_count = INT_MAX;
  for( i = 0;  i < KheTimeSetTimeCount(ts);  i++ )
  {
    t = KheTimeSetTime(ts, i);
    ti = HaArray(gs->time_info, KheTimeIndex(t));
    HaArrayForEach(ti->root_task_sets, rts, j)
    {
      rts_count = KheResourceGroupResourceCount(rts->domain);
      if( rts_count >= res_count )
	break;  ** later ones have larger domains **
      if( KheRootTaskSetAcceptable(rts, check_ignore, nsubset_ts, disjoint_ts,
	    ass_to_rts) )
      {
	*res = rts;
	res_count = rts_count;
      }
    }
  }
  return *res != NULL;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheFindLeaderRootTaskSet(KHE_RCG_SOLVER gs,                         */
/*    KHE_RESOURCE_GROUP domain, bool allow_single,                          */
/*    KHE_ROOT_TASK_SET *res)                                                */
/*                                                                           */
/*  Find a leader root task set from the times of gs->tbg_in_ts.  This must  */
/*                                                                           */
/*    * Not be marked unsuitable for providing leader tasks                  */
/*    * Cover at least one of the times of gs->tbg_in_ts                     */
/*    * Have domain equal to domain, unless domain == NULL                   */
/*    * Not cover all of the times of gs->tbg_in_ts, unless allow_single     */
/*    * Not cover any of the times of gs->tbg_out_ts                         */
/*                                                                           */
/*  Among all such root task sets, give first priority to those with an      */
/*  initial assignment, and second priority to increasing domain size.       */
/*                                                                           */
/*  If a suitable leader root task set can be found, set *res to it and      */
/*  return true.  Otherwise set *res to NULL and return false.               */
/*                                                                           */
/*****************************************************************************/

static bool KheFindLeaderRootTaskSet(KHE_RCG_SOLVER gs,
  KHE_RESOURCE_GROUP domain, bool allow_single, KHE_ROOT_TASK_SET *res)
{
  int i, j;  KHE_TIME t;  KHE_TIME_INFO ti;  KHE_ROOT_TASK_SET rts;
  if( DEBUG5 )
  {
    fprintf(stderr, "          [ KheFindLeaderRootTaskSet(in ");
    KheTimeSetDebug(gs->tbg_in_ts, 2, -1, stderr);
    fprintf(stderr, ", out ");
    KheTimeSetDebug(gs->tbg_out_ts, 2, -1, stderr);
    fprintf(stderr, ", %s, %s)\n",
      domain == NULL ? "-" : KheResourceGroupId(domain),
      bool_show(allow_single));
  }
  *res = NULL;
  for( i = 0;  i < KheTimeSetTimeCount(gs->tbg_in_ts);  i++ )
  {
    t = KheTimeSetTime(gs->tbg_in_ts, i);
    ti = HaArray(gs->time_info, KheTimeIndex(t));
    HaArrayForEach(ti->root_task_sets, rts, j)
      if( !rts->ignore_for_leader &&
	  (domain == NULL || KheResourceGroupEqual(domain, rts->domain)) &&
	  (allow_single || !KheTimeSetSubset(gs->tbg_in_ts, rts->time_set)) &&
	  KheTimeSetDisjoint(gs->tbg_out_ts, rts->time_set) &&
	  (*res == NULL || KheRootTaskSetLT(rts, *res)) )
	*res = rts;
  }
  if( DEBUG5 || DEBUG8 )
  {
    fprintf(stderr, "            leader:  ");
    if( *res != NULL )
      KheRootTaskSetDebug(*res, 0, stderr);
    else
      fprintf(stderr, "none\n");
    if( DEBUG5 )
      fprintf(stderr, "          ] KheFindLeaderRootTaskSet returning %s\n",
	bool_show(*res != NULL));
  }
  return *res != NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheFindFollowerRootTaskSet(KHE_RCG_SOLVER gs,                       */
/*    KHE_ROOT_TASK_SET root_rts, KHE_RESOURCE init_asst,                    */
/*    KHE_ROOT_TASK_SET *res, KHE_ROOT_TASK_SET *problem_rts)                */
/*                                                                           */
/*  Assuming that gs->tbg_in_ts contains at least one time, find a           */
/*  follower task set from the first time of gs->tbg_in_ts.  This must       */
/*                                                                           */
/*    * Have init_asst (possibly NULL) for its initial assignment            */
/*    * Cover at least one of the times of gs->tbg_in_ts                     */
/*    * Not cover any of the times of gs->tbg_out_ts                         */
/*    * Be movable to root_rts                                               */
/*                                                                           */
/*  If a suitable follower root task set can be found, set *res to it and    */
/*  return true.  Otherwise set *res to NULL and return false.               */
/*                                                                           */
/*  If we come upon a root task set with a non-NULL initial assignment       */
/*  that is equal to init_asst, we really want to choose it.  If we can't    */
/*  choose it, we set *problem_rts to it to tell the caller of the problem.  */
/*                                                                           */
/*****************************************************************************/

static bool KheFindFollowerRootTaskSet(KHE_RCG_SOLVER gs,
  KHE_ROOT_TASK_SET root_rts, KHE_RESOURCE init_asst,
  KHE_ROOT_TASK_SET *res, KHE_ROOT_TASK_SET *problem_rts)
{
  int i;  KHE_TIME t;  KHE_TIME_INFO ti;  KHE_ROOT_TASK_SET rts;
  KHE_TASK root_task;
  t = KheTimeSetTime(gs->tbg_in_ts, 0);
  ti = HaArray(gs->time_info, KheTimeIndex(t));
  root_task =  KheTaskSetTask(root_rts->tasks, 0);
  HaArrayForEach(ti->root_task_sets, rts, i)
  {
    if( rts->init_asst == init_asst )
    {
      if( KheTimeSetDisjoint(gs->tbg_out_ts, rts->time_set) &&
	  KheTaskMoveCheck(KheTaskSetTask(rts->tasks, 0), root_task) )
      {
	if( DEBUG8 )
	{
	  fprintf(stderr, "            follower: ");
	  KheRootTaskSetDebug(rts, 0, stderr);
	}
	return *res = rts, true;
      }
      if( init_asst != NULL )
      {
	/* this is a problem, we really want to assign rts in this case */
	if( DEBUG8 )
	{
	  fprintf(stderr, "            problem: ");
	  KheRootTaskSetDebug(rts, 0, stderr);
	}
	*problem_rts = rts;
      }
    }
  }
  if( DEBUG8 )
    fprintf(stderr, "            follower: NULL\n");
  return *res = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheInitializeTimeSets(KHE_RCG_SOLVER gs, KHE_TIME_SET in_ts,        */
/*    KHE_TIME_SET out_ts)                                                   */
/*                                                                           */
/*  Initialize gs->tbg_in_ts and gs->tbg_out_ts to in_ts and out_ts plus,    */
/*  for each time t in in_ts, the times of the common frame time group of    */
/*  t other than t itself.                                                   */
/*                                                                           */
/*****************************************************************************/

static void KheInitializeTimeSets(KHE_RCG_SOLVER gs, KHE_TIME_SET in_ts,
  KHE_TIME_SET out_ts)
{
  int i;  KHE_TIME t;  KHE_TIME_GROUP tg;
  KheTimeSetCopyElements(gs->tbg_in_ts, in_ts);
  KheTimeSetCopyElements(gs->tbg_out_ts, out_ts);
  for( i = 0;  i < KheTimeSetTimeCount(in_ts);  i++ )
  {
    t = KheTimeSetTime(in_ts, i);
    tg = KheFrameTimeTimeGroup(gs->common_frame, t);
    KheTimeSetAddTimeGroup(gs->tbg_out_ts, tg);
    KheTimeSetDeleteTime(gs->tbg_out_ts, t);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheUpdateTimeSets(KHE_RCG_SOLVER gs, KHE_ROOT_TASK_SET rts)         */
/*                                                                           */
/*  Update gs->tbg_in_ts and gs->tbg_out_ts to reflect the inclusion of      */
/*  rts->times.  That is, remove rts->times from gs->tbg_in_ts, and add      */
/*  every day covered by the times of rts->times from gs->tbg_out_ts.        */
/*                                                                           */
/*****************************************************************************/

static void KheUpdateTimeSets(KHE_RCG_SOLVER gs, KHE_ROOT_TASK_SET rts)
{
  int i;  KHE_TIME t;  KHE_TIME_GROUP tg;
  KheTimeSetDifference(gs->tbg_in_ts, rts->time_set);
  for( i = 0;  i < KheTimeSetTimeCount(rts->time_set);  i++ )
  {
    t = KheTimeSetTime(rts->time_set, i);
    tg = KheFrameTimeTimeGroup(gs->common_frame, t);
    KheTimeSetAddTimeGroup(gs->tbg_out_ts, tg);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTimeBasedGrouping(KHE_RCG_SOLVER gs, KHE_TIME_SET in_ts,         */
/*    KHE_TIME_SET out_ts, KHE_RESOURCE_GROUP leader_domain, int max_num,    */
/*    bool allow_single, KHE_TASK *test_task, int *groups_made)              */
/*                                                                           */
/*  Try to ensure that there are up to max_num groups of tasks such that     */
/*  in each group, each time of in_ts is covered, and no time of out_ts is   */
/*  covered.  Existing groups with these properties count towards max_num.   */
/*  Return true if at least one group was made, and set *groups_made to      */
/*  the number of groups actually made.                                      */
/*                                                                           */
/*  If leader_domain != NULL, then only tasks whose domain is equal to       */
/*  leader_domain are considered for being leader tasks.                     */
/*                                                                           */
/*  If allow_single is true, it is permissible for a group to contain        */
/*  just a single task.                                                      */
/*                                                                           */
/*  If test_task is non-NULL, then if at least one group was made, set       */
/*  *test_task to its root task, otherwise set *test_task to NULL.  Also     */
/*  do not update the data structure and do not print debug output saying    */
/*  that a group was made, because this is just a test and it will be        */
/*  unmade shortly after this function returns.                              */
/*                                                                           */
/*  Time set in_ts will never contain two times from the same time group     */
/*  of the common frame, and time sets in_ts and out_ts will never overlap.  */
/*  It does not matter whether the grouped tasks run at times outside in_ts  */
/*  and out_ts or not.  However, no two of them may run at the same time,    */
/*  or at two times from the same time group of the common frame.            */
/*                                                                           */
/*  It is acceptable for in_ts to contain only one time.  This can produce   */
/*  no true grouping, but it is allowed because combinatorial grouping       */
/*  needs to know whether assigning a single time is cost-free.              */
/*                                                                           */
/*****************************************************************************/

static bool KheTimeBasedGrouping(KHE_RCG_SOLVER gs, KHE_TIME_SET in_ts,
  KHE_TIME_SET out_ts, KHE_RESOURCE_GROUP leader_domain, int max_num,
  bool allow_single, KHE_TASK *test_task, int *groups_made, char *what)
{
  KHE_ROOT_TASK_SET root_rts, rts, problem_rts;  KHE_TIME t;  KHE_TIME_INFO ti;
  KHE_TASK root_task, task;  int i, j, count, groups_count;
  if( DEBUG5 || DEBUG8 )
  {
    fprintf(stderr, "        [ KheTimeBasedGrouping(gs, in ");
    KheTimeSetDebug(in_ts, 2, -1, stderr);
    fprintf(stderr, ", out ");
    KheTimeSetDebug(out_ts, 2, -1, stderr);
    fprintf(stderr, ", %s, max %d)\n",
      leader_domain == NULL ? "-" : KheResourceGroupId(leader_domain), max_num);
    if( DEBUG5 )
      for( i = 0;  i < KheTimeSetTimeCount(in_ts);  i++ )
      {
	t = KheTimeSetTime(in_ts, i);
	ti = HaArray(gs->time_info, KheTimeIndex(t));
	KheTimeInfoDebug(ti, 10, stderr);
      }
  }
  /* ***
  HnAssert(KheTimeSetTimeCount(in_ts) >= 2,
    "KheTimeBasedGrouping: in_ts has less than 2 elements");
  *** */
  root_task = NULL;
  *groups_made = 0;
  HaArrayClear(gs->tbg_ignore_for_leader_set);
  KheInitializeTimeSets(gs, in_ts, out_ts);
  root_rts = NULL;  /* keep compiler happy */
  while( max_num > 0 &&
    KheFindLeaderRootTaskSet(gs, leader_domain, allow_single, &root_rts) )
  {
    /* update time sets for root_rts */
    KheUpdateTimeSets(gs, root_rts);

    /* find the follower root task sets */
    HaArrayClear(gs->tbg_rts_set);
    problem_rts = NULL;
    if( root_rts->init_asst != NULL )
    {
      while( KheTimeSetTimeCount(gs->tbg_in_ts) > 0 &&
	  KheFindFollowerRootTaskSet(gs, root_rts, root_rts->init_asst, &rts,
	    &problem_rts) )
      {
	HaArrayAddLast(gs->tbg_rts_set, rts);
	KheUpdateTimeSets(gs, rts);
      }
    }
    while( KheTimeSetTimeCount(gs->tbg_in_ts) > 0 &&
	KheFindFollowerRootTaskSet(gs, root_rts, NULL, &rts, &problem_rts) )
    {
      HaArrayAddLast(gs->tbg_rts_set, rts);
      KheUpdateTimeSets(gs, rts);
    }

    if( problem_rts == NULL && KheTimeSetTimeCount(gs->tbg_in_ts) == 0 )
    {
      /* success; find the number of groups to make */
      groups_count = max_num;
      count = KheTaskSetTaskCount(root_rts->tasks);
      if( count < groups_count )
	groups_count = count;
      HaArrayForEach(gs->tbg_rts_set, rts, i)
      {
	count = KheTaskSetTaskCount(rts->tasks);
	if( count < groups_count )
	  groups_count = count;
      }
      HnAssert(groups_count >= 1, "KheTimeBasedGrouping internal error 1");

      /* make the groups */
      if( test_task != NULL )
      {
	/* grouping for testing only; don't update data structures */
	HnAssert(max_num == 1, "KheTimeBasedGrouping internal error 2");
	root_task = KheTaskSetLast(root_rts->tasks);
	HaArrayForEach(gs->tbg_rts_set, rts, j)
	{
	  task = KheTaskSetLast(rts->tasks);
	  if( !KheTaskMove(task, root_task) )
	    HnAbort("KheTimeBasedGrouping internal error 3");
	}
	*test_task = root_task;
	*groups_made += 1;
      }
      else
      {
	/* grouping for real, including updating data structures */
	for( i = 0;  i < groups_count;  i++ )
	{
	  root_task = KheRootTaskSetLastAndDelete(root_rts);
	  HaArrayForEach(gs->tbg_rts_set, rts, j)
	  {
	    task = KheRootTaskSetLastAndDelete(rts);
	    if( !KheTaskMove(task, root_task) )
	      HnAbort("KheTimeBasedGrouping internal error 4");
	    gs->soln_changed = true;
	    if( gs->task_set != NULL )
	      KheTaskSetAddTask(gs->task_set, task);
	  }
	  if( DEBUG3 )
	  {
	    fprintf(stderr, "  %s made grouped task (%s) ", what,
	      KheResourceGroupId(KheTaskDomain(root_task)));
	    KheTaskDebug(root_task, 2, 0, stderr);
	  }
	  KheGroupSolverAddTask(gs, root_task);
	  *groups_made += 1;
	}
      }

      /* update max_num */
      max_num -= groups_count;
    }
    else
    {
      /* problem with root_rts, so ignore it and retry */
      root_rts->ignore_for_leader = true;
      HaArrayAddLast(gs->tbg_ignore_for_leader_set, root_rts);
    }
    KheInitializeTimeSets(gs, in_ts, out_ts);
    root_rts = NULL;  /* keep compiler happy */
  }

  /* all done */
  /* NB some of the root task sets visited here may be on the free list */
  HaArrayForEach(gs->tbg_ignore_for_leader_set, rts, i)
    rts->ignore_for_leader = false;
  if( DEBUG5 || DEBUG8 )
    fprintf(stderr, "        ] KheTimeBasedGrouping returning %s\n",
      root_task != NULL ? "true" : "false");
  return root_task != NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "properties of constraints"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int KheLimitBusyTimesConstraintResourceTypeCount(                        */
/*    KHE_LIMIT_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE rt)               */
/*                                                                           */
/*  Return the number of resources of type rt monitored by c.                */
/*                                                                           */
/*****************************************************************************/

static int KheLimitBusyTimesConstraintResourceTypeCount(
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE rt)
{
  int i, count;  KHE_RESOURCE_GROUP rg;  KHE_RESOURCE r;
  count = 0;
  for( i = 0;  i < KheLimitBusyTimesConstraintResourceCount(c);  i++ )
  {
    r = KheLimitBusyTimesConstraintResource(c, i);
    if( KheResourceResourceType(r) == rt )
      count += 1;
  }
  for( i = 0;  i < KheLimitBusyTimesConstraintResourceGroupCount(c);  i++ )
  {
    rg = KheLimitBusyTimesConstraintResourceGroup(c, i);
    if( KheResourceGroupResourceType(rg) == rt )
      count += KheResourceGroupResourceCount(rg);
  }
  return count;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheClusterBusyTimesConstraintResourceTypeCount(                      */
/*    KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE rt)             */
/*                                                                           */
/*  Return the number of resources of type rt monitored by c.                */
/*                                                                           */
/*****************************************************************************/

static int KheClusterBusyTimesConstraintResourceTypeCount(
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE rt)
{
  int i, count;  KHE_RESOURCE_GROUP rg;  KHE_RESOURCE r;
  count = 0;
  for( i = 0;  i < KheClusterBusyTimesConstraintResourceCount(c);  i++ )
  {
    r = KheClusterBusyTimesConstraintResource(c, i);
    if( KheResourceResourceType(r) == rt )
      count += 1;
  }
  for( i = 0;  i < KheClusterBusyTimesConstraintResourceGroupCount(c);  i++ )
  {
    rg = KheClusterBusyTimesConstraintResourceGroup(c, i);
    if( KheResourceGroupResourceType(rg) == rt )
      count += KheResourceGroupResourceCount(rg);
  }
  return count;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintResourceTypeCount(                  */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE_TYPE rt)         */
/*                                                                           */
/*  Return the number of resources of type rt monitored by c.                */
/*                                                                           */
/*****************************************************************************/

static int KheLimitActiveIntervalsConstraintResourceTypeCount(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE_TYPE rt)
{
  int i, count;  KHE_RESOURCE_GROUP rg;  KHE_RESOURCE r;
  count = 0;
  for( i = 0;  i < KheLimitActiveIntervalsConstraintResourceCount(c);  i++ )
  {
    r = KheLimitActiveIntervalsConstraintResource(c, i);
    if( KheResourceResourceType(r) == rt )
      count += 1;
  }
  for( i = 0;  i < KheLimitActiveIntervalsConstraintResourceGroupCount(c); i++ )
  {
    rg = KheLimitActiveIntervalsConstraintResourceGroup(c, i);
    if( KheResourceGroupResourceType(rg) == rt )
      count += KheResourceGroupResourceCount(rg);
  }
  return count;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDemandAtTime(KHE_EVENT_TIMETABLE_MONITOR etm,                     */
/*    KHE_RESOURCE_TYPE rt, KHE_TIME t)                                      */
/*                                                                           */
/*  Return the total demand for resources of type rt at time t.              */
/*                                                                           */
/*****************************************************************************/

static int KheDemandAtTime(KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_RESOURCE_TYPE rt, KHE_TIME t)
{
  int j, k, res;  KHE_MEET meet;  KHE_TASK task;  KHE_RESOURCE r;
  res = 0;
  for( j = 0;  j < KheEventTimetableMonitorTimeMeetCount(etm, t);  j++ )
  {
    meet = KheEventTimetableMonitorTimeMeet(etm, t, j);
    for( k = 0;  k < KheMeetTaskCount(meet);  k++ )
    {
      task = KheMeetTask(meet, k);
      if( KheTaskResourceType(task) == rt &&
	(KheTaskIsPreassigned(task, &r) || KheTaskNonAssignmentHasCost(task)) )
	res++;
    }
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDemandAtTimeGroup(KHE_EVENT_TIMETABLE_MONITOR etm,                */
/*    KHE_RESOURCE_TYPE rt, KHE_TIME_GROUP tg)                               */
/*                                                                           */
/*  Return the total demand for resources of type rt in the times of tg.     */
/*                                                                           */
/*****************************************************************************/

static int KheDemandAtTimeGroup(KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_RESOURCE_TYPE rt, KHE_TIME_GROUP tg)
{
  int i, res;  KHE_TIME t;
  res = 0;
  for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
  {
    t = KheTimeGroupTime(tg, i);
    res += KheDemandAtTime(etm, rt, t);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDemandAtClusterConstraint(KHE_EVENT_TIMETABLE_MONITOR etm,        */
/*    KHE_RESOURCE_TYPE rt, KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, int offset) */
/*                                                                           */
/*  Return the total demand for resources of type rt in the time groups      */
/*  of cluster busy times constraint c at offset.                            */
/*                                                                           */
/*****************************************************************************/

static int KheDemandAtClusterConstraint(KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_RESOURCE_TYPE rt, KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, int offset)
{
  int i, res;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  res = 0;
  for( i = 0;  i < KheClusterBusyTimesConstraintTimeGroupCount(c);  i++ )
  {
    tg = KheClusterBusyTimesConstraintTimeGroup(c, i, offset, &po);
    res += KheDemandAtTimeGroup(etm, rt, tg);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheClusterFirstAndLastIndex(KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c,    */
/*    int offset, KHE_FRAME frame, int *first_index, int *last_index)        */
/*                                                                           */
/*  Return the first and last indexes in frame where the time groups         */
/*  intersect with the time groups of c at offset.  If there is no           */
/*  intersection, return an empty interval (*first_index > *last_index).     */
/*                                                                           */
/*****************************************************************************/

static bool KheClusterFirstAndLastIndex(KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c,
  int offset, KHE_FRAME frame, int *first_index, int *last_index)
{
  int i;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  KHE_TIME t, earliest_t, latest_t;

  /* find the earliest and latest times; if none, return false */
  earliest_t = latest_t = NULL;
  for( i = 0;  i < KheClusterBusyTimesConstraintTimeGroupCount(c);  i++ )
  {
    tg = KheClusterBusyTimesConstraintTimeGroup(c, i, offset, &po);
    if( KheTimeGroupTimeCount(tg) > 0 )
    {
      t = KheTimeGroupTime(tg, 0);
      if( earliest_t == NULL || KheTimeIndex(t) < KheTimeIndex(earliest_t) )
        earliest_t = t;
      t = KheTimeGroupTime(tg, KheTimeGroupTimeCount(tg) - 1);
      if( latest_t == NULL || KheTimeIndex(t) > KheTimeIndex(latest_t) )
        latest_t = t;
    }
  }
  if( earliest_t == NULL )
    return *first_index = 0, *last_index = -1, false;

  /* all good, return the indexes */
  return *first_index = KheFrameTimeIndex(frame, earliest_t),
    *last_index = KheFrameTimeIndex(frame, latest_t), true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitFirstAndLastIndex(KHE_LIMIT_BUSY_TIMES_CONSTRAINT c,        */
/*    int offset, KHE_FRAME frame, int *first_index, int *last_index)        */
/*                                                                           */
/*  Return the first and last indexes in frame where the time groups         */
/*  intersect with the time groups of c at offset.  If there is no           */
/*  intersection, return an empty interval (*first_index > *last_index).     */
/*                                                                           */
/*****************************************************************************/

static bool KheLimitFirstAndLastIndex(KHE_LIMIT_BUSY_TIMES_CONSTRAINT c,
  int offset, KHE_FRAME frame, int *first_index, int *last_index)
{
  int i;  KHE_TIME_GROUP tg;  KHE_TIME t, earliest_t, latest_t;

  /* find the earliest and latest times; if none, return false */
  earliest_t = latest_t = NULL;
  for( i = 0;  i < KheLimitBusyTimesConstraintTimeGroupCount(c);  i++ )
  {
    tg = KheLimitBusyTimesConstraintTimeGroup(c, i, offset);
    if( KheTimeGroupTimeCount(tg) > 0 )
    {
      t = KheTimeGroupTime(tg, 0);
      if( earliest_t == NULL || KheTimeIndex(t) < KheTimeIndex(earliest_t) )
        earliest_t = t;
      t = KheTimeGroupTime(tg, KheTimeGroupTimeCount(tg) - 1);
      if( latest_t == NULL || KheTimeIndex(t) > KheTimeIndex(latest_t) )
        latest_t = t;
    }
  }
  if( earliest_t == NULL )
    return *first_index = 0, *last_index = -1, false;

  /* all good, return the indexes */
  return *first_index = KheFrameTimeIndex(frame, earliest_t),
    *last_index = KheFrameTimeIndex(frame, latest_t), true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheClusterConstraintHasNonSingletonTimeGroup(                       */
/*    KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc)                                */
/*                                                                           */
/*  Return true if cbtc has at least one time group with >=2 elements.       */
/*                                                                           */
/*****************************************************************************/

static bool KheClusterConstraintHasNonSingletonTimeGroup(
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc)
{
  int i;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  for( i = 0;  i < KheClusterBusyTimesConstraintTimeGroupCount(cbtc);  i++ )
  {
    tg = KheClusterBusyTimesConstraintTimeGroup(cbtc, i, 0, &po);
    if( KheTimeGroupTimeCount(tg) >= 2 )
      return true;
  }
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheClusterConstraintSuitsCombinationElimination(                    */
/*    KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE rt, int *lim)   */
/*                                                                           */
/*  Return true if c is suited to combination elimination:  if its time      */
/*  groups are all positive and it has a non-trivial maximum limit, or its   */
/*  time groups are all negative and it has a non-trivial minimum limit.     */
/*                                                                           */
/*  If c is suitable, also set *lim to a maximum limit on the number of      */
/*  its busy (not active) time groups.  This will be c's maximum limit if    */
/*  its time groups are all positive, or the number of time groups minus     */
/*  the minimum limit if its time groups are all negative.                   */
/*                                                                           */
/*  Also check that c's time groups are disjoint, that it has at least one   */
/*  non-singleton time group, and that it covers all resources of type rt.   */
/*                                                                           */
/*****************************************************************************/

static bool KheClusterConstraintSuitsCombinationElimination(
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE rt, int *lim)
{
  int max_lim, min_lim, count, rt_count;
  rt_count = KheResourceTypeResourceCount(rt);
  if( KheClusterBusyTimesConstraintTimeGroupsDisjoint(c) &&
      KheClusterConstraintHasNonSingletonTimeGroup(c) &&
      KheClusterBusyTimesConstraintResourceTypeCount(c, rt) == rt_count )
  {
    max_lim = KheClusterBusyTimesConstraintMaximum(c);
    min_lim = KheClusterBusyTimesConstraintMinimum(c);
    count = KheClusterBusyTimesConstraintTimeGroupCount(c);
    if( KheClusterBusyTimesConstraintAllPositive(c) && max_lim < count )
      return *lim = max_lim, true;
    else if( KheClusterBusyTimesConstraintAllNegative(c) && min_lim > 0 )
      return *lim = count - min_lim, true;
    else
      return *lim = -1, false;
  }
  else
    return *lim = -1, false;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "eliminating combinations based on supply and demand"          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int IntCmpIncreasing(const void *t1, const void *t2)                     */
/*                                                                           */
/*  Comparison fuunction for sorting an array of integers into increasing    */
/*  order.                                                                   */
/*                                                                           */
/*****************************************************************************/

static int IntCmpIncreasing(const void *t1, const void *t2)
{
  int val1 = * (int *) t1;
  int val2 = * (int *) t2;
  return val1 - val2;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheEliminateCombinationsForConstraint(KHE_RCG_SOLVER gs,            */
/*    KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, int lim)                          */
/*                                                                           */
/*  Eliminate combinations based on supply and demand within the times of    */
/*  c, a suitable cluster busy times monitor with limit lim.                 */
/*                                                                           */
/*****************************************************************************/
#define offset_count KheClusterBusyTimesConstraintAppliesToOffsetCount

static void KheEliminateCombinationsForConstraint(KHE_RCG_SOLVER gs,
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, int lim)
{
  KHE_TIME_GROUP c_tg, f_tg, prev_f_tg;
  int offset, i, j, count, sum, fi, li, supply, demand;  KHE_POLARITY po;
  bool f_tg_overlaps, prev_f_tg_overlaps;  KHE_TIME_GROUP_INFO tgi;
  if( DEBUG6 )
  {
    fprintf(stderr, "[ KheEliminateCombinationsForConstraint(gs, c, %d), c:\n",
      lim);
    KheClusterBusyTimesConstraintDebug(c, 3, 2, stderr);
  }

  /* for each time group of c, find the number of frame time groups it */
  /* intersects with.  Also record in ec_time_group_info those time groups */
  /* which will be linked with previous time groups if demand allows */
  for( offset = 0;  offset < offset_count(c);  offset++ )
  {
    HaArrayClear(gs->ec_time_group_info);
    HaArrayClear(gs->ec_intersects);
    KheClusterFirstAndLastIndex(c, offset, gs->common_frame, &fi, &li);
    for( i = 0;  i < KheClusterBusyTimesConstraintTimeGroupCount(c);  i++ )
    {
      c_tg = KheClusterBusyTimesConstraintTimeGroup(c, i, offset, &po);
      count = 0;
      prev_f_tg = NULL;
      prev_f_tg_overlaps = false;
      for( j = fi;  j <= li;  j++ )
      {
	f_tg = KheFrameTimeGroup(gs->common_frame, j);
	f_tg_overlaps = !KheTimeGroupDisjoint(c_tg, f_tg);
	if( f_tg_overlaps )
	  count++;
	if( f_tg_overlaps && prev_f_tg_overlaps )
	{
	  /* if we later find that D(T) >= S(T), then if prev_f_tg is */
	  /* busy, then f_tg must be busy as well, because they both  */
	  /* overlap the same time group of the monitor */
	  if( DEBUG6 )
	  {
	    fprintf(stderr, "  provisionally linked time groups ");
	    KheTimeGroupDebug(prev_f_tg, 1, -1, stderr);
	    fprintf(stderr, " and ");
	    KheTimeGroupDebug(f_tg, 1, -1, stderr);
	    fprintf(stderr, "\n");
	  }

	  /* record the linkage provisionally */
	  tgi = HaArray(gs->time_group_info, j);
	  HaArrayAddLast(gs->ec_time_group_info, tgi);
	}
	prev_f_tg = f_tg;
	prev_f_tg_overlaps = f_tg_overlaps;
      }
      HaArrayAddLast(gs->ec_intersects, count);
    }

    /* get the supply:  the largest lim intersect counts, times resources */
    HaArraySort(gs->ec_intersects, &IntCmpIncreasing);
    count = HaArrayCount(gs->ec_intersects);
    sum = 0;
    for( i = 1;  i <= lim;  i++ )
      sum += HaArray(gs->ec_intersects, count - i);
    supply = sum * KheResourceTypeResourceCount(gs->resource_type);

    /* get the demand */
    demand = KheDemandAtClusterConstraint(gs->etm, gs->resource_type,c,offset);

    /* if demand >= supply, provisional linkages become definite */
    if( demand >= supply )
      HaArrayForEach(gs->ec_time_group_info, tgi, i)
	tgi->linked_to_prev = true;
    if( DEBUG6 )
      fprintf(stderr, "  offset %d, demand %d, supply %d\n",
	offset, demand, supply);
  }
  if( DEBUG6 )
    fprintf(stderr, "] KheEliminateCombinationsForConstraint returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheEliminateCombinations(KHE_RCG_SOLVER gs)                         */
/*                                                                           */
/*  Eliminate combinations based on supply and demand.                       */
/*                                                                           */
/*  NB the condition ClusterBusyTimesConstraintTimeGroupsDisjoint(cbtc) is   */
/*  not really required, but it always seems to hold, and it makes the       */
/*  calculation of demand much easier.                                       */
/*                                                                           */
/*****************************************************************************/
static void KheGroupByResourceConstraintsSolverDebug(KHE_RCG_SOLVER gs,
  int indent, FILE *fp);

static void KheEliminateCombinations(KHE_RCG_SOLVER gs)
{
  int i, lim;  KHE_CONSTRAINT c;  KHE_RESOURCE r;  KHE_RESOURCE_TYPE rt;
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;
  rt = gs->resource_type;
  if( DEBUG6 )
    fprintf(stderr, "[ KheEliminateCombinations(%s)\n", KheResourceTypeId(rt));
  r = KheResourceTypeResource(rt, 0);  /* must exist, see call on this fn */
  for( i = 0;  i < KheResourceConstraintCount(r);  i++ )
  {
    c = KheResourceConstraint(r, i);
    if( KheConstraintTag(c) == KHE_CLUSTER_BUSY_TIMES_CONSTRAINT_TAG )
    {
      cbtc = (KHE_CLUSTER_BUSY_TIMES_CONSTRAINT) c;
      if( KheClusterConstraintSuitsCombinationElimination(cbtc, rt, &lim) )
	KheEliminateCombinationsForConstraint(gs, cbtc, lim);
    }
  }
  if( DEBUG6 )
  {
    fprintf(stderr, "  at end of KheEliminateCombinations:\n");
    KheGroupByResourceConstraintsSolverDebug(gs, 2, stderr);
    fprintf(stderr, "] KheEliminateCombinations returning\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "combinatorial grouping"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheMonitorSuitsCombinatorialGrouping(KHE_MONITOR m,                 */
/*    KHE_RCG_SOLVER gs)                                                     */
/*                                                                           */
/*  Return true if m suits combinatorial grouping:  if it is a cluster busy  */
/*  times or limit busy times monitor whose times all lie in the current     */
/*  interval, derived from a constraint that applies to all resources.       */
/*                                                                           */
/*****************************************************************************/

static bool KheMonitorSuitsCombinatorialGrouping(KHE_MONITOR m,
  KHE_RCG_SOLVER gs)
{
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm;
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT lbtc;  KHE_LIMIT_BUSY_TIMES_MONITOR lbtm;
  int offset, fi, li, count;  KHE_RESOURCE_TYPE rt;  KHE_FRAME frame;
  rt = gs->resource_type;
  count = KheResourceTypeResourceCount(rt);
  frame = gs->common_frame;
  switch( KheMonitorTag(m) )
  {
    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      cbtm = (KHE_CLUSTER_BUSY_TIMES_MONITOR) m;
      cbtc = KheClusterBusyTimesMonitorConstraint(cbtm);
      offset = KheClusterBusyTimesMonitorOffset(cbtm);
      return KheClusterBusyTimesConstraintResourceTypeCount(cbtc, rt) == count
	&& KheClusterFirstAndLastIndex(cbtc, offset, frame, &fi, &li) &&
	fi >= gs->cg_first_index && li <= gs->cg_last_index;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      lbtm = (KHE_LIMIT_BUSY_TIMES_MONITOR) m;
      lbtc = KheLimitBusyTimesMonitorConstraint(lbtm);
      offset = KheLimitBusyTimesMonitorOffset(lbtm);
      return KheLimitBusyTimesConstraintResourceTypeCount(lbtc, rt) == count
	&& KheLimitFirstAndLastIndex(lbtc, offset, frame, &fi, &li) &&
	fi >= gs->cg_first_index && li <= gs->cg_last_index;

    default:

      return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTraceHasZeroCost(KHE_TRACE trace, KHE_RCG_SOLVER gs)             */
/*                                                                           */
/*  Return true if the monitors that show up on trace all have zero cost.    */
/*  Actually we are only interested in monitors which are cluster busy       */
/*  times and limit busy times monitors derived from constraints that        */
/*  apply to all resources, and whose times lie between gs->cg_first_index   */
/*  and gs->cg_last_index in the common frame.                               */
/*                                                                           */
/*****************************************************************************/

static bool KheTraceHasZeroCost(KHE_RCG_SOLVER gs)
{
  int i;  KHE_MONITOR m;  bool cost_increased, suits;
  for( i = 0;  i < KheTraceMonitorCount(gs->cg_trace);  i++ )
  {
    m = KheTraceMonitor(gs->cg_trace, i);
    cost_increased =
      (KheMonitorCost(m) > KheTraceMonitorInitCost(gs->cg_trace, i));
    suits = KheMonitorSuitsCombinatorialGrouping(m, gs);
    if( DEBUG4 )
    {
      fprintf(stderr, "        %c%c %.5f -> %.5f ",
        cost_increased ? 'C' : ' ', suits ? 'S' : ' ',
	KheCostShow(KheTraceMonitorInitCost(gs->cg_trace, i)),
	KheCostShow(KheMonitorCost(m)));
      KheMonitorDebug(m, 1, 0, stderr);
    }
    if( cost_increased && suits )
      return false;
  }
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheCombinatorialGroupingAssign(KHE_RCG_SOLVER gs, int curr_index,   */
/*    bool prev_tg_assigned)                                                 */
/*                                                                           */
/*  Carry out the part of the solve that assigns all combinations of tasks   */
/*  and checks which ones are successful.                                    */
/*                                                                           */
/*  If prev_tg_assigned, then the previous time group (at curr_index - 1)    */
/*  has been assigned something, or is off the left end.                     */
/*                                                                           */
/*    cg_first_index          The first index of the area being searched     */
/*    cg_last_index           The last index of the area being searched      */
/*    cg_in_ts                The times that have so far been included       */
/*    cg_out_ts               The times that have so far been excluded       */
/*                                                                           */
/*    cg_success_count        Number of successes                            */
/*    cg_success_in_ts        If success, the times included                 */
/*    cg_success_out_ts       If success, the times excluded                 */
/*                                                                           */
/*    cg_trace                Temporary variable holding a trace object      */
/*                                                                           */
/*****************************************************************************/

static void KheCombinatorialGroupingAssign(KHE_RCG_SOLVER gs, int curr_index,
  bool prev_tg_assigned)
{
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_TASK root_task;  KHE_MARK mark;
  KHE_RESOURCE_GROUP rg;  KHE_RESOURCE r;  int i, groups_made;
  KHE_TIME_GROUP_INFO tgi;  bool res;
  if( curr_index > gs->cg_last_index )
  {
    /* time sets are complete; try for a grouped task now */
    mark = KheMarkBegin(gs->soln);
    KheTraceBegin(gs->cg_trace);
    if( DEBUG4 )
    {
      fprintf(stderr, "      [ testing in_ts ");
      KheTimeSetDebug(gs->cg_in_ts, 2, -1, stderr);
      fprintf(stderr, " and out_ts ");
      KheTimeSetDebug(gs->cg_out_ts, 2, -1, stderr);
      fprintf(stderr, ":\n");
    }
    if( KheTimeBasedGrouping(gs, gs->cg_in_ts, gs->cg_out_ts, NULL, 1, true,
	  &root_task, &groups_made, NULL) )
    {
      /* assign root_task and check the result of tracing it */
      if( DEBUG4 )
      {
	fprintf(stderr, "        root_task ");
	KheTaskDebug(root_task, 2, 0, stderr);
      }
      rg = KheTaskDomain(root_task);
      if( KheResourceGroupResourceCount(rg) > 0 )
      {
	r = KheTaskAsstResource(root_task);
	if( r != NULL )
	  res = KheTraceHasZeroCost(gs);
	else
	{
	  r = KheResourceGroupResource(rg, 0);
	  res = KheTaskMoveResource(root_task, r) && KheTraceHasZeroCost(gs);
	}
	if( res )
	{
	  gs->cg_success_count++;
	  KheTimeSetCopyElements(gs->cg_success_in_ts, gs->cg_in_ts);
	  KheTimeSetCopyElements(gs->cg_success_out_ts, gs->cg_out_ts);
	  if( DEBUG4 )
	    fprintf(stderr, "      ] success\n");
	}
	else
	{
	  if( DEBUG4 )
	    fprintf(stderr, "      ] failure (no assign, or non-zero cost)\n");
	}
      }
      else
      {
	if( DEBUG4 )
	  fprintf(stderr, "      ] failure (no resources)\n");
      }
    }
    else
    {
      if( DEBUG4 )
	fprintf(stderr, "      ] failure (no group)\n");
    }
    KheTraceEnd(gs->cg_trace);
    KheMarkEnd(mark, true);
  }
  else
  {
    /* try assigning each time from the current time group */
    tg = KheFrameTimeGroup(gs->common_frame, curr_index);
    for( i = 0; gs->cg_success_count < 2 && i < KheTimeGroupTimeCount(tg); i++ )
    {
      t = KheTimeGroupTime(tg, i);
      KheTimeSetAddTime(gs->cg_in_ts, t);
      KheCombinatorialGroupingAssign(gs, curr_index + 1, true);
      KheTimeSetDeleteTime(gs->cg_in_ts, t);
    }

    /* try not assigning any time from the current time group */
    if( gs->cg_success_count < 2 )
    {
      tgi = HaArray(gs->time_group_info, curr_index);
      if( prev_tg_assigned && tgi->linked_to_prev )
      {
	if( DEBUG6 )
	{
	  fprintf(stderr, "  due to linkage, omitting non-assignment of ");
	  KheTimeGroupDebug(tg, 1, 0, stderr);
	}
      }
      else
      {
	KheTimeSetAddTimeGroup(gs->cg_out_ts, tg);
	KheCombinatorialGroupingAssign(gs, curr_index + 1, false);
	KheTimeSetDeleteTimeGroup(gs->cg_out_ts, tg);
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDoCombinatorialGroupingForInterval(KHE_RCG_SOLVER gs,            */
/*    int first_index, int last_index)                                       */
/*                                                                           */
/*  Solve gs for the interval first_index .. last_index, and if there was    */
/*  exactly one success, do time-based grouping for that set of times.       */
/*                                                                           */
/*****************************************************************************/

static void KheDoCombinatorialGroupingForInterval(KHE_RCG_SOLVER gs,
  int first_index, int last_index)
{
  int i, groups_made;  bool success;  KHE_TIME_GROUP tg;  KHE_TIME t;
  if( DEBUG2 )
  {
    fprintf(stderr, "  [ KheDoCombinatorialGroupingForInterval(gs, ");
    tg = KheFrameTimeGroup(gs->common_frame, first_index);
    KheTimeGroupDebug(tg, 1, -1, stderr);
    fprintf(stderr, " .. ");
    tg = KheFrameTimeGroup(gs->common_frame, last_index);
    KheTimeGroupDebug(tg, 1, -1, stderr);
    fprintf(stderr, ")\n");
  }
  gs->cg_first_index = first_index;
  gs->cg_last_index = last_index;

  /* try one solve for each time in the first time group, but then */
  /* start all over again and have another go if any were successful */
  tg = KheFrameTimeGroup(gs->common_frame, first_index);
  do
  {
    success = false;
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      /* initialize the solve fields other than gm */
      KheTimeSetClear(gs->cg_in_ts);
      KheTimeSetClear(gs->cg_out_ts);
      gs->cg_success_count = 0;
      KheTimeSetClear(gs->cg_success_in_ts);
      KheTimeSetClear(gs->cg_success_out_ts);

      /* solve for t */
      t = KheTimeGroupTime(tg, i);
      if( DEBUG4 )
	fprintf(stderr, "    [ combinations starting with %s:\n", KheTimeId(t));
      KheTimeSetAddTime(gs->cg_in_ts, t);
      KheCombinatorialGroupingAssign(gs, first_index + 1, true);
      KheTimeSetDeleteTime(gs->cg_in_ts, t);
      if( DEBUG4 )
	fprintf(stderr, "    ] (%d successes)\n", gs->cg_success_count);

      /* if exactly one success, we have succeeded */
      if( gs->cg_success_count == 1 )
      {
	/* have a single successful assignment, so group tasks */
	if( DEBUG2 )
	{
	  fprintf(stderr, "  KheDoCombinatorialGroupingForInterval found ");
	  KheTimeSetDebug(gs->cg_success_in_ts, 2, 0, stderr);
	}
	if( KheTimeBasedGrouping(gs, gs->cg_success_in_ts,
	      gs->cg_success_out_ts, NULL, INT_MAX, false, NULL,
	      &groups_made, "combinatorial grouping") )
	  success = true;
	/* KheGroupTasksCombinatorial(gs, gs->success_ts); */
      }
    }
  } while( success );

  /* KheGroupMonitorDelete(gs->cg_group_monitor); */
  /* gs->cg_group_monitor = NULL; */
  if( DEBUG2 )
    fprintf(stderr, "  ] KheDoCombinatorialGroupingForInterval returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDoCombinatorialGrouping(KHE_RCG_SOLVER gs)                       */
/*                                                                           */
/*  Carry out combinatorial grouping for resource type gs->resource_type.    */
/*                                                                           */
/*****************************************************************************/

static void KheDoCombinatorialGrouping(KHE_RCG_SOLVER gs)
{
  int len, first_index, days;

  /* solve for each subsequence of len days */
  /* ***
  if( DEBUG2 )
    KheGroupByResourceConstraintsSolverDebug(gs, 2, stderr);
  *** */
  days = KheFrameTimeGroupCount(gs->common_frame);
  for( len = gs->cg_max_days;  len >= 2;  len-- )
    for( first_index = 0;  first_index <= days - len;  first_index++ )
      KheDoCombinatorialGroupingForInterval(gs, first_index,
	first_index + len - 1);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "profile grouping"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsMonitorSuitsProfileGrouping(                 */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR m, KHE_RESOURCE_TYPE rt)            */
/*                                                                           */
/*  Return true if m suits profile grouping:  if its constraint applies to   */
/*  all resources of type rt, and its time groups are positive singletons.   */
/*                                                                           */
/*****************************************************************************/

static bool KheLimitActiveIntervalsMonitorSuitsProfileGrouping(
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR m, KHE_RESOURCE_TYPE rt)
{
  int i, rt_count;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c;

  /* there must be at least two time groups */
  if( KheLimitActiveIntervalsMonitorTimeGroupCount(m) < 2 )
    return false;

  /* its time groups must be positive singletons */
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(m);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(m, i, &po);
    if( KheTimeGroupTimeCount(tg) != 1 || po != KHE_POSITIVE )
      return false;
  }

  /* it must monitor all resources of type rt */
  rt_count = KheResourceTypeResourceCount(rt);
  c = KheLimitActiveIntervalsMonitorConstraint(m);
  if( KheLimitActiveIntervalsConstraintResourceTypeCount(c, rt) < rt_count )
    return false;

  /* all OK */
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsMonitorsCanGroupTogether(                    */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim1,                              */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim2)                              */
/*                                                                           */
/*  Return true if laim1 and laim2 can be handled together, because they     */
/*  have equal time groups, one has a maximum limit only, and the other      */
/*  has a minimum limit only.                                                */
/*                                                                           */
/*****************************************************************************/

static bool KheLimitActiveIntervalsMonitorsCanGroupTogether(
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim1,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim2)
{
  int i;  KHE_TIME_GROUP tg1, tg2;  KHE_POLARITY po1, po2;
  bool has_min1, has_min2, has_max1, has_max2;

  /* time groups must be equal */
  if( KheLimitActiveIntervalsMonitorTimeGroupCount(laim1) !=
      KheLimitActiveIntervalsMonitorTimeGroupCount(laim2) )
    return false;
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim1);  i++ )
  {
    tg1 = KheLimitActiveIntervalsMonitorTimeGroup(laim1, i, &po1);
    tg2 = KheLimitActiveIntervalsMonitorTimeGroup(laim2, i, &po2);
    if( !KheTimeGroupEqual(tg1, tg2) )
      return false;
  }

  /* one must have a min limit only, the other a max limit only */
  has_min1 = KheLimitActiveIntervalsMonitorMinimum(laim1) > 0;
  has_min2 = KheLimitActiveIntervalsMonitorMinimum(laim2) > 0;
  has_max1 = KheLimitActiveIntervalsMonitorMaximum(laim1) <
    KheLimitActiveIntervalsMonitorTimeGroupCount(laim1);
  has_max2 = KheLimitActiveIntervalsMonitorMaximum(laim2) <
    KheLimitActiveIntervalsMonitorTimeGroupCount(laim2);
  if( (has_min1 && !has_max1) && (!has_min2 && has_max2) )
    return true;
  if( (!has_min1 && has_max1) && (has_min2 && !has_max2) )
    return true;
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsMonitorSetTimes(                             */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR m, KHE_TIME_SET ts)                 */
/*                                                                           */
/*  Set the times of ts to the times of the time groups of m.                */
/*                                                                           */
/*****************************************************************************/

static void KheLimitActiveIntervalsMonitorSetTimes(
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR m, KHE_TIME_SET ts)
{
  int i, tg_count;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  KheTimeSetClear(ts);
  tg_count = KheLimitActiveIntervalsMonitorTimeGroupCount(m);
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(m, i, &po);
    HnAssert(KheTimeGroupTimeCount(tg) == 1 && po == KHE_POSITIVE,
      "KheLimitActiveIntervalsMonitorSetTimes internal error");
    KheTimeSetAddTimeGroup(ts, tg);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheProfileGroupingBegin(KHE_RCG_SOLVER gs)                          */
/*                                                                           */
/*  Begin profile grouping for gs->pg_monitor_ts.                            */
/*                                                                           */
/*****************************************************************************/

static void KheProfileGroupingBegin(KHE_RCG_SOLVER gs)
{
  int i, j;  KHE_TIME t;  KHE_ROOT_TASK_SET rts;  KHE_TIME_INFO ti;
  gs->pg_active = true;
  for( i = 0;  i < KheTimeSetTimeCount(gs->pg_monitor_ts);  i++ )
  {
    /* initialize the profile grouping fields of ti and its root task sets */
    t = KheTimeSetTime(gs->pg_monitor_ts, i);
    ti = HaArray(gs->time_info, KheTimeIndex(t));
    ti->pg_active = true;
    ti->pg_profile = 0;
    HaArrayForEach(ti->root_task_sets, rts, j)
    {
      /* determine the profile grouping status of rts (may already be done) */
      if( rts->pg_status == KHE_PG_INACTIVE_UNMONITORED )
        rts->pg_status = KheRootTaskSetStatus(rts);

      /* add rts's contribution to ti's profile */
      switch( rts->pg_status )
      {
	case KHE_PG_INACTIVE_MAXIMAL:

	  /* maximal tasks do not contribute to the profile */
	  break;

	case KHE_PG_ACTIVE:

	  /* add to profile */
	  ti->pg_profile += KheTaskSetTaskCount(rts->tasks);
	  break;

	case KHE_PG_INACTIVE_UNMONITORED:
	default:

	  /* there are no unmonitored tasks at gs->pg_monitor_ts times */
	  HnAbort("KheProfileGroupingBegin internal error");
	  break;
      }
    }
    if( DEBUG9 )
      fprintf(stderr, "  profile at %s: %d\n", KheTimeId(t), ti->pg_profile);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheProfileGroupingEnd(KHE_RCG_SOLVER gs)                            */
/*                                                                           */
/*  End profile grouping for gs->pg_monitor_ts.                              */
/*                                                                           */
/*****************************************************************************/

static void KheProfileGroupingEnd(KHE_RCG_SOLVER gs)
{
  int i, j;  KHE_TIME t;  KHE_ROOT_TASK_SET rts;  KHE_TIME_INFO ti;
  for( i = 0;  i < KheTimeSetTimeCount(gs->pg_monitor_ts);  i++ )
  {
    /* clear out the profile grouping fields of ti and its root task sets */
    t = KheTimeSetTime(gs->pg_monitor_ts, i);
    ti = HaArray(gs->time_info, KheTimeIndex(t));
    ti->pg_active = false;
    ti->pg_profile = 0;
    HaArrayForEach(ti->root_task_sets, rts, j)
      rts->pg_status = KHE_PG_INACTIVE_UNMONITORED;
  }
  gs->pg_active = false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheAddTime(KHE_RCG_SOLVER gs, int index, KHE_TIME_SET ts)           */
/*                                                                           */
/*  Add the i'th element of gs->pg_monitor_ts to ts, but only if it exists.  */
/*                                                                           */
/*****************************************************************************/

static void KheAddTime(KHE_RCG_SOLVER gs, int index, KHE_TIME_SET ts)
{
  KHE_TIME t;
  if( 0 <= index && index < KheTimeSetTimeCount(gs->pg_monitor_ts) )
  {
    t = KheTimeSetTime(gs->pg_monitor_ts, index);
    KheTimeSetAddTime(ts, t);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheProfileGroupingSetInTimeSet(KHE_RCG_SOLVER gs,                   */
/*    int first_index, int last_index)                                       */
/*                                                                           */
/*  Set gs->pg_in_ps to include all the times of gs->pg_monitor_ts from      */
/*  first_index to last_index inclusive, except out-of-range ones.           */
/*                                                                           */
/*****************************************************************************/

static void KheProfileGroupingSetInTimeSet(KHE_RCG_SOLVER gs,
  int first_index, int last_index)
{
  int i;
  KheTimeSetClear(gs->pg_in_ts);
  for( i = first_index;  i <= last_index;  i++ )
    KheAddTime(gs, i, gs->pg_in_ts);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheProfileGroupingSetOutTimeSet(KHE_RCG_SOLVER gs, int index1,      */
/*   int index2)                                                             */
/*                                                                           */
/*  Set gs->pg_out_ts to exclude the times at index1 and index2 from         */
/*  gs->pg_monitor_ts.  Either index may be out of range, in which case      */
/*  it is ignored.                                                           */
/*                                                                           */
/*****************************************************************************/

static void KheProfileGroupingSetOutTimeSet(KHE_RCG_SOLVER gs, int index1,
 int index2)
{
  KheTimeSetClear(gs->pg_out_ts);
  KheAddTime(gs, index1, gs->pg_out_ts);
  KheAddTime(gs, index2, gs->pg_out_ts);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheTasksSatisfyingInAndOut(KHE_RCG_SOLVER gs,                        */
/*    KHE_TIME_SET in_ts, KHE_TIME_SET out_ts)                               */
/*                                                                           */
/*  Return the number of tasks that cover all of in_ts and none of out_ts.   */
/*                                                                           */
/*****************************************************************************/

static int KheTasksSatisfyingInAndOut(KHE_RCG_SOLVER gs,
  KHE_TIME_SET in_ts, KHE_TIME_SET out_ts)
{
  KHE_TIME t;  KHE_TIME_INFO ti;  KHE_ROOT_TASK_SET rts;  int i, res;
  t = KheTimeSetTime(in_ts, 0);
  ti = HaArray(gs->time_info, KheTimeIndex(t));
  res = 0;
  HaArrayForEach(ti->root_task_sets, rts, i)
    if( KheTimeSetSubset(in_ts, rts->time_set) &&
	KheTimeSetDisjoint(out_ts, rts->time_set) )
      res += KheTaskSetTaskCount(rts->tasks);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheActiveTasksWithDomain(KHE_TIME_INFO ti, KHE_RESOURCE_GROUP rg)    */
/*                                                                           */
/*  Return the number of pg-active tasks in ti's root task sets with         */
/*  domain rg.                                                               */
/*                                                                           */
/*****************************************************************************/

static int KheActiveTasksWithDomain(KHE_TIME_INFO ti, KHE_RESOURCE_GROUP rg)
{
  int i, res;  KHE_ROOT_TASK_SET rts;
  if( DEBUG11 )
    fprintf(stderr, "[ KheActiveTasksWithDomain(%s, %s)\n",
      KheTimeId(ti->time), KheResourceGroupId(rg));
  res = 0;
  HaArrayForEach(ti->root_task_sets, rts, i)
  {
    if( DEBUG11 )
      KheRootTaskSetDebug(rts, 2, stderr);
    if( rts->pg_status == KHE_PG_ACTIVE &&
	KheResourceGroupEqual(rts->domain, rg) )
      res += KheTaskSetTaskCount(rts->tasks);
  }
  if( DEBUG11 )
    fprintf(stderr, "] KheActiveTasksWithDomain returning %d\n", res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDoProfile(KHE_RCG_SOLVER gs, KHE_TIME_INFO ti,                   */
/*    KHE_TIME_INFO adj_ti, int max_num)                                     */
/*                                                                           */
/*  Do profile grouping starting or ending at ti.  The number of groups      */
/*  to make is max_num.                                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDoProfile(KHE_RCG_SOLVER gs, KHE_TIME_INFO ti,
  KHE_TIME_INFO adj_ti, int max_num)
{
  KHE_RESOURCE_GROUP domain;
  int i, pos, domain_count, groups_made;  KHE_ROOT_TASK_SET rts;

  /* build the list of domains and counts */
  if( DEBUG10 )
    fprintf(stderr, "[ KheDoProfile(gs, ti %s, adj_ti %s, max_num %d)\n",
      KheTimeId(ti->time), adj_ti == NULL ? "-" : KheTimeId(adj_ti->time),
      max_num);
  HaArrayClear(gs->pg_domains);
  HaArrayClear(gs->pg_domain_counts);
  HaArrayForEach(ti->root_task_sets, rts, i)
    if( !HaArrayContains(gs->pg_domains, rts->domain, &pos) )
    {
      /* NB if there is history we won't reach this point anyway */
      domain_count = KheActiveTasksWithDomain(ti, rts->domain) -
	(adj_ti == NULL ? 0 : KheActiveTasksWithDomain(adj_ti, rts->domain));
      HaArrayAddLast(gs->pg_domains, rts->domain);
      HaArrayAddLast(gs->pg_domain_counts, domain_count);
      if( DEBUG10 )
	fprintf(stderr, "  adding (%s, %d)\n",
	  KheResourceGroupId(rts->domain), domain_count);
    }

  /* traverse the list and do the grouping */
  for( i = 0;  i < HaArrayCount(gs->pg_domains) && max_num > 0;  i++ )
  {
    domain = HaArray(gs->pg_domains, i);
    domain_count = HaArray(gs->pg_domain_counts, i);
    if( domain_count > 0 )
    {
      KheTimeBasedGrouping(gs, gs->pg_in_ts, gs->pg_out_ts, domain,
	min(domain_count, max_num), false, NULL, &groups_made,
	"profile grouping");
      max_num -= groups_made;
      if( DEBUG10 )
	fprintf(stderr, "  made %d groups for %s", groups_made,
	  KheResourceGroupId(rts->domain));
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDoProfileGroupingForMonitors(KHE_RCG_SOLVER gs,                  */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim1,                              */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim2)                              */
/*                                                                           */
/*  Carry out profile grouping based on laim1 and (optionally) laim2.        */
/*                                                                           */
/*****************************************************************************/

static void KheDoProfileGroupingForMonitors(KHE_RCG_SOLVER gs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim1,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim2)
{
  int tcount, prev_profile, next_profile, i, demand, supply;
  KHE_TIME t;  KHE_TIME_INFO ti, prev_ti, next_ti;

  if( DEBUG9 )
  {
    fprintf(stderr, "[ KheDoProfileGroupingForMonitors(gs, laim1, %s)\n",
      laim2 != NULL ? "laim2" : "-");
    KheLimitActiveIntervalsMonitorDebug(laim1, 1, 2, stderr);
    if( laim2 != NULL )
      KheLimitActiveIntervalsMonitorDebug(laim2, 1, 2, stderr);
  }

  /* get min_limit, max_limit, history_before, and history_after */
  gs->pg_min_limit = KheLimitActiveIntervalsMonitorMinimum(laim1);
  gs->pg_max_limit = KheLimitActiveIntervalsMonitorMaximum(laim1);
  gs->pg_history_before = KheLimitActiveIntervalsMonitorHistoryBefore(laim1)> 0;
  gs->pg_history_after  = KheLimitActiveIntervalsMonitorHistoryAfter(laim1) > 0;
  if( laim2 != NULL )
  {
    if( KheLimitActiveIntervalsMonitorMinimum(laim2) > gs->pg_min_limit )
      gs->pg_min_limit = KheLimitActiveIntervalsMonitorMinimum(laim2);
    if( KheLimitActiveIntervalsMonitorMaximum(laim2) < gs->pg_max_limit )
      gs->pg_max_limit = KheLimitActiveIntervalsMonitorMaximum(laim2);
    if( KheLimitActiveIntervalsMonitorHistoryBefore(laim2) > 0 )
      gs->pg_history_before = true;
    if( KheLimitActiveIntervalsMonitorHistoryAfter(laim2)  > 0 )
      gs->pg_history_after = true;
  }

  if( gs->pg_min_limit >= 2 )
  {
    /* build the time set for laim1 and initialize profile grouping fields */
    KheLimitActiveIntervalsMonitorSetTimes(laim1, gs->pg_monitor_ts);
    tcount = KheTimeSetTimeCount(gs->pg_monitor_ts);
    KheProfileGroupingBegin(gs);

    /* look for places where sequences start */
    prev_ti = NULL;
    prev_profile = gs->pg_history_before ? INT_MAX : 0;
    for( i = 0;  i <= tcount - 2;  i++ )
    {
      /* find the time and time info for i */
      t = KheTimeSetTime(gs->pg_monitor_ts, i);
      ti = HaArray(gs->time_info, KheTimeIndex(t));

      if( ti->pg_profile > prev_profile )
      {
	/* ti->pg_profile - prev_profile sequences must start at i */
	demand = ti->pg_profile - prev_profile;
        KheProfileGroupingSetInTimeSet(gs, i, i + gs->pg_min_limit - 1);
	KheProfileGroupingSetOutTimeSet(gs, i - 1, i + gs->pg_max_limit);
	supply = KheTasksSatisfyingInAndOut(gs, gs->pg_in_ts, gs->pg_out_ts);
	if( supply < demand )
	{
	  if( DEBUG9 )
	    fprintf(stderr, "  grouping from %s (supply %d, demand %d - %d):\n",
	      KheTimeId(t), supply, ti->pg_profile, prev_profile);
          KheDoProfile(gs, ti, prev_ti, demand - supply);
	  /* ***
	  KheTimeBasedGrouping(gs, gs->pg_in_ts, gs->pg_out_ts, NULL,
	    demand - supply, false, NULL, &groups_made);
	  *** */
	}
      }

      /* set prev_profile for next iteration */
      prev_ti = ti;
      prev_profile = ti->pg_profile;
    }

    /* look for places where sequences end */
    next_ti = NULL;
    next_profile = gs->pg_history_after ? INT_MAX : 0;
    for( i = tcount - 1;  i >= 1;  i-- )
    {
      /* find the time and time info for i */
      t = KheTimeSetTime(gs->pg_monitor_ts, i);
      ti = HaArray(gs->time_info, KheTimeIndex(t));

      if( ti->pg_profile > next_profile )
      {
	/* ti->pg_profile - next_profile sequences must end at i */
	demand = ti->pg_profile - next_profile;
        KheProfileGroupingSetInTimeSet(gs, i - gs->pg_min_limit + 1, i);
	KheProfileGroupingSetOutTimeSet(gs, i + 1, i - gs->pg_max_limit);
	supply = KheTasksSatisfyingInAndOut(gs, gs->pg_in_ts, gs->pg_out_ts);
	if( supply < demand )
	{
	  if( DEBUG9 )
	    fprintf(stderr, "  grouping to %s (supply %d, demand %d - %d):\n",
	      KheTimeId(t), supply, ti->pg_profile, next_profile);
          KheDoProfile(gs, ti, next_ti, demand - supply);
	  /* ***
	  KheTimeBasedGrouping(gs, gs->pg_in_ts, gs->pg_out_ts, NULL,
	    demand - supply, false, NULL, &groups_made);
	  *** */
	}
      }

      /* set next_profile for next iteration */
      next_ti = ti;
      next_profile = ti->pg_profile;
    }
    KheProfileGroupingEnd(gs);
  }

  if( DEBUG9 )
    fprintf(stderr, "] KheDoProfileGroupingForMonitors returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDoProfileGrouping(KHE_RCG_SOLVER gs)                             */
/*                                                                           */
/*  Carry out profile grouping for gs.                                       */
/*                                                                           */
/*****************************************************************************/

static void KheDoProfileGrouping(KHE_RCG_SOLVER gs)
{
  KHE_MONITOR m;  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laimi, laimj;
  int i, j;  KHE_RESOURCE r;  KHE_RESOURCE_TYPE rt;

  /* find suitable limit active intervals monitors */
  HaArrayClear(gs->pg_monitors);
  rt = gs->resource_type;
  r = KheResourceTypeResource(rt, 0);
  for( i = 0;  i < KheSolnResourceMonitorCount(gs->soln, r);  i++ )
  {
    m = KheSolnResourceMonitor(gs->soln, r, i);
    if( KheMonitorTag(m) == KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG )
    {
      laimi = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
      if( KheLimitActiveIntervalsMonitorSuitsProfileGrouping(laimi, rt) )
      {
	if( DEBUG7 )
	{
	  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c;
	  c = KheLimitActiveIntervalsMonitorConstraint(laimi);
	  fprintf(stderr, "  found suitable monitor, constraint %s\n",
	    KheConstraintId((KHE_CONSTRAINT) c));
	}
	HaArrayAddLast(gs->pg_monitors, laimi);
      }
    }
  }

  /* find pairs of suitable monitors and do the job for each pair */
  for( i = 0;  i < HaArrayCount(gs->pg_monitors);  i++ )
  {
    laimi = HaArray(gs->pg_monitors, i);
    for( j = i + 1;  j < HaArrayCount(gs->pg_monitors);  j++ )
    {
      laimj = HaArray(gs->pg_monitors, j);
      if( KheLimitActiveIntervalsMonitorsCanGroupTogether(laimi, laimj) )
      {
	KheDoProfileGroupingForMonitors(gs, laimi, laimj);
	HaArrayDeleteAndPlug(gs->pg_monitors, j);  /* order matters! */
	HaArrayDeleteAndPlug(gs->pg_monitors, i);  /* order matters! */
	i--;
	break;
      }
    }
  }

  /* find single suitable monitors and do the job for them */
  HaArrayForEach(gs->pg_monitors, laimi, i)
    KheDoProfileGroupingForMonitors(gs, laimi, NULL);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "main function"                                                */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheGroupByResourceConstraintsForResourceType(KHE_SOLN soln,         */
/*    KHE_RESOURCE_TYPE rt, int max_days, bool resource_invariant,           */
/*    KHE_FRAME common_frame, KHE_EVENT_TIMETABLE_MONITOR etm,               */
/*    KHE_TASK_SET ts, HA_ARENA a)                                           */
/*                                                                           */
/*  Carry out group by resource constraints for resource type rt.  There     */
/*  is no point in doing anything if there are fewer than two resources.     */
/*                                                                           */
/*****************************************************************************/

static bool KheGroupByResourceConstraintsForResourceType(KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt, int max_days, bool resource_invariant,
  KHE_FRAME common_frame, KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_TASK_SET ts, HA_ARENA a)
{
  KHE_RCG_SOLVER gs;
  if( KheResourceTypeResourceCount(rt) > 1 )
  {
    gs = KheGroupByResourceConstraintsSolverMake(soln, rt, max_days,
      resource_invariant, common_frame, etm, ts, a);
    KheEliminateCombinations(gs);
    KheDoCombinatorialGrouping(gs);
    KheDoProfileGrouping(gs);
    return gs->soln_changed;
  }
  else
    return false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheGroupByResourceConstraints(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,  */
/*    KHE_OPTIONS options, KHE_TASK_SET ts)                                  */
/*                                                                           */
/*  Group the tasks of soln of type rt by resource constraints.  If rt is    */
/*  NULL, do this for each of the resource types of soln's instance.         */
/*                                                                           */
/*  If ts is non-NULL, add every task that is given an assignment to ts.     */
/*                                                                           */
/*****************************************************************************/

bool KheGroupByResourceConstraints(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,
  KHE_OPTIONS options, KHE_TASK_SET ts)
{
  int max_days, i;  bool res, resource_invariant;  KHE_INSTANCE ins;
  HA_ARENA a;  KHE_FRAME common_frame;  KHE_EVENT_TIMETABLE_MONITOR etm;

  /* boilerplate */
  ins = KheSolnInstance(soln);
  if( DEBUG1 )
    fprintf(stderr, "[ KheGroupByResourceConstraints(%s %s, -)\n",
      KheInstanceId(ins), rt == NULL ? "all types" : KheResourceTypeId(rt));

  /* return immediately if option rs_group_by_frame_off is true */
  if( KheOptionsGetBool(options, "rs_group_by_rc_off", false) )
  {
    if( DEBUG1 )
      fprintf(stderr,"] KheGroupByResourceConstraints returning false (off)\n");
    return false;
  }

  /* get the other options used by this function */
  max_days = KheOptionsGetInt(options, "rs_group_by_rc_max_days", 3);
  resource_invariant = KheOptionsGetBool(options, "rs_invariant", false);
  common_frame = KheOptionsFrame(options, "gs_common_frame", soln);
  etm = (KHE_EVENT_TIMETABLE_MONITOR)
    KheOptionsGetObject(options, "gs_event_timetable_monitor", NULL);
  HnAssert(etm != NULL, "KheGroupByResourceConstraints: required "
    "option gs_event_timetable_monitor not present");

  /* do grouping by resource constraints for each resource type */
  a = KheSolnArenaBegin(soln);
  res = false;
  if( rt != NULL )
  {
    if( KheGroupByResourceConstraintsForResourceType(soln, rt, max_days,
	resource_invariant, common_frame, etm, ts, a) )
      res = true;
  }
  else for( i = 0;  i < KheInstanceResourceTypeCount(ins);  i++ )
  {
    rt = KheInstanceResourceType(ins, i);
    if( KheGroupByResourceConstraintsForResourceType(soln, rt, max_days,
	resource_invariant, common_frame, etm, ts, a) )
      res = true;
  }
  KheSolnArenaEnd(soln, a);

  /* all done */
  if( DEBUG1 )
    fprintf(stderr, "] KheGroupByResourceConstraints returning %s\n",
      bool_show(res));
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_FRAMED_MONITOR - a monitor, plus a frame saying what it covers       */
/*                                                                           */
/*  Phase 0 applies when elim_lim >= 0.                                      */
/*                                                                           */
/*****************************************************************************/

/* ***
typedef struct khe_framed_monitor_rec
{
  KHE_MONITOR		monitor;
  int			first_index;
  int			last_index;
  int			elim_lim;
  HA_ARRAY_INT		elim_intersects;
  int			elim_supply;
  int			elim_demand;
} *KHE_FRAMED_MONITOR;

typedef HA_ARRAY(KHE_FRAMED_MONITOR) ARRAY_KHE_FRAMED_MONITOR;
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_SET - a set of times (like a time group, but without baggage)   */
/*                                                                           */
/*****************************************************************************/

/* *** moved to platform, khe_time_set.c
typedef HA_ARRAY(KHE_TIME) ARRAY_KHE_TIME;

typedef struct khe_time_set_rec
{
  ARRAY_KHE_TIME		times;
} *KHE_TIME_SET;
*** */



/*****************************************************************************/
/*                                                                           */
/*  KHE_DOMAIN_TASK_SET - a set of tasks, broken by time, with one domain    */
/*                                                                           */
/*****************************************************************************/

/* ***
typedef struct khe_domain_task_set_rec {
  KHE_RESOURCE_GROUP			domain;		** the shared domain **
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR	monitor;	** the monitor       **
  int					min_limit;	** min limit         **
  int					max_limit;	** min limit         **
  ARRAY_KHE_TASK_SET			task_sets;	** one per laim tg   **
} *KHE_DOMAIN_TASK_SET;

typedef HA_ARRAY(KHE_DOMAIN_TASK_SET) ARRAY_KHE_DOMAIN_TASK_SET;
*** */

/*****************************************************************************/
/*                                                                           */
/*  Submodule "domain task sets"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DOMAIN_TASK_SET KheDomainTaskSetMake(KHE_GROUP_BY_FRAME_SOLVER gfs,  */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR monitor, KHE_RESOURCE_GROUP domain, */
/*    int min_limit, int max_limit)                                          */
/*                                                                           */
/*  Make a new domain task set with these attributes.                        */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DOMAIN_TASK_SET KheDomainTaskSetMake(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR monitor, KHE_RESOURCE_GROUP domain,
  int min_limit, int max_limit)
{
  KHE_DOMAIN_TASK_SET res;  int i, tg_count;  KHE_TASK_SET ts;
  HaMake(res, gfs->arena);
  res->domain = domain;
  res->monitor = monitor;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  HaArrayInit(res->task_sets, gfs->arena);
  tg_count = KheLimitActiveIntervalsMonitorTimeGroupCount(monitor);
  for( i = 0;  i < tg_count;  i++ )
  {
    ts = KheTaskSetMake(gfs->soln);
    HaArrayAddLast(res->task_sets, ts);
  }
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetAddTask(KHE_DOMAIN_TASK_SET dts, KHE_TASK task,     */
/*    int index)                                                             */
/*                                                                           */
/*  Add task to dts, at the appropriate index.                               */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDomainTaskSetAddTask(KHE_DOMAIN_TASK_SET dts, KHE_TASK task,
  int index)
{
  KHE_TASK_SET ts;
  HnAssert(0 <= index && index < HaArrayCount(dts->task_sets),
    "KheDomainTaskSetAddTask: index (%d) out of range (0 .. %d)",
    index, HaArrayCount(dts->task_sets) - 1);
  ts = HaArray(dts->task_sets, index);
  HnAssert(ts != NULL, "KheDomainTaskSetAddTask internal error");
  KheTaskSetAddTask(ts, task);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDomainTaskSetFindPlateau(KHE_DOMAIN_TASK_SET dts,                */
/*    int *start_pos, int *end_pos)                                          */
/*                                                                           */
/*  If dts contains a plateau, return the first one.                         */
/*                                                                           */
/*  Writing Num(i) for the number of tasks at position i, a plateau          */
/*  is a pair of indexes (start_pos, end_pos) satisfying                     */
/*                                                                           */
/*    Num(start_pos - 1) < Num(start_pos) &&                                 */
/*    Num(start_pos) == Num(start_pos + 1) == ... == Num(end_pos) &&         */
/*    Num(end_pos) > Num(end_pos + 1)                                        */
/*                                                                           */
/*  This means it is possible to group tasks at start_pos ... end_pos,       */
/*  although whether grouping will actually be possible depends on the       */
/*  length of the plateau (end_pos - start_pos + 1) as well.                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDomainTaskSetFindPlateau(KHE_DOMAIN_TASK_SET dts,
  int *start_pos, int *end_pos)
{
  KHE_TASK_SET ts1, ts2;  int i, prev_up;
  prev_up = -1;
  for( i = 0;  i < HaArrayCount(dts->task_sets) - 1;  i++ )
  {
    ts1 = HaArray(dts->task_sets, i);
    ts2 = HaArray(dts->task_sets, i + 1);
    if( KheTaskSetTaskCount(ts1) < KheTaskSetTaskCount(ts2) )
      prev_up = i + 1;
    else if( KheTaskSetTaskCount(ts1) > KheTaskSetTaskCount(ts2) )
    {
      if( prev_up != -1 )
	return *start_pos = prev_up, *end_pos = i, true;
    }
  }
  return *start_pos = -1, *end_pos = -1, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetFindTaskRange(KHE_DOMAIN_TASK_SET dts,              */
/*    KHE_TASK task, int start_pos, int *end_pos)                            */
/*                                                                           */
/*  Assuming that task is present at start_pos, set *end_pos to the end      */
/*  point of its range.                                                      */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDomainTaskSetFindTaskRange(KHE_DOMAIN_TASK_SET dts,
  KHE_TASK task, int start_pos, int *end_pos)
{
  int i;
  for( i = start_pos + 1;  i <= 
  {
    ts = HaArray(dts->task_sets, j);
    if( !KheTaskSetContainsTask(ts, task, &pos) )
      break;
  }

}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDomainTaskSetContainsGroupableTask(KHE_DOMAIN_TASK_SET dts,      */
/*    int *start_pos, int end_pos, KHE_TASK *res)                            */
/*                                                                           */
/*  Search for a task that covers *start_pos ... end_pos, update them        */
/*  and return the task.                                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDomainTaskSetContainsGroupableTask(KHE_DOMAIN_TASK_SET dts,
  int *start_pos, int end_pos, KHE_TASK *res)
{
  KHE_TASK_SET ts;  int i, j, k, pos;  KHE_TASK task;
  ts = HaArray(dts->task_sets, *start_pos);
  for( i = 0;  i < KheTaskSetTaskCount(ts);  i++ )
  {
    ** see whether task is groupable **
    task = KheTaskSetTask(ts, i);

    ** set *start_pos ... j - 1 to the range covered by task **
    for( j = *start_pos + 1;  j <= end_pos;  j++ )
    {
      ts = HaArray(dts->task_sets, j);
      if( !KheTaskSetContainsTask(ts, task, &pos) )
	break;
    }

    ** if this range equals the total duration of task, return it **
    if( j - *start_pos == KheTaskTotalDuration(task) )
    {
      for( k = *start_pos;  k < j;  k++ )
      {
	ts = HaArray(dts->task_sets, k);
	KheTaskSetDeleteTask(ts, task);
      }
      return *start_pos = j, *res = task, true;
    }
  }
  return *start_pos = -1, *res = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetGroup(KHE_DOMAIN_TASK_SET dts,                      */
/*    int start_pos, int end_pos, KHE_GROUP_BY_FRAME_SOLVER gfs)             */
/*                                                                           */
/*  Remove one task from each task set in this range and group them.  Be     */
/*  careful to take account of existing groupings:  if a task is part of     */
/*  a group that extends outside this range, don't use it; and if it         */
/*  covers several indexes within the range, treat those as all done.        */
/*                                                                           */
/*  Any tasks that don't work get removed anyway.                            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDomainTaskSetGroup(KHE_DOMAIN_TASK_SET dts,
  int start_pos, int end_pos, KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  KHE_TASK first_task, task;  bool ok, starting;
  first_task = NULL;  starting = true;
  while( KheDomainTaskSetContainsGroupableTask(dts, &start_pos, end_pos, &task))
  {
    if( first_task == NULL )
      first_task = task;
    else
    {
      ok = KheTaskAssign(task, first_task);
      if( ok )
      {
	if( DEBUG3 )
	{
	  if( starting )
	  {
	    fprintf(stderr, "    KheGroupByFrame plateau grouping tasks {");
	    KheTaskDebug(first_task, 2, -1, stderr);
	    starting = false;
	  }
	  fprintf(stderr, ", ");
	  KheTaskDebug(task, 2, -1, stderr);
	}
	gfs->soln_changed = true;
	if( gfs->task_set != NULL )
	  KheTaskSetAddTask(gfs->task_set, task);
      }
    }
  }
  if( DEBUG3 && !starting )
    fprintf(stderr, "}\n");
}
*** */


/* *** old version that does not take existing groupings into account
static void KheDomainTaskSetGroup(KHE_DOMAIN_TASK_SET dts,
  int start_pos, int end_pos, KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  KHE_TASK_SET ts;  int i;  KHE_TASK first_task, task;  bool ok;
  ts = HaArray(dts->task_sets, start_pos);
  first_task = KheTaskSetTask(ts, KheTaskSetTaskCount(ts) - 1);
  if( DEBUG3 )
  {
    fprintf(stderr, "    KheGroupByFrame plateau grouping tasks {");
    KheTaskDebug(first_task, 1, -1, stderr);
  }
  for( i = start_pos + 1;  i <= end_pos;  i++ )
  {
    ts = HaArray(dts->task_sets, i);
    task = KheTaskSetTask(ts, KheTaskSetTaskCount(ts) - 1);
    if( task == first_task )
      ok = true;
    else
    {
      ok = KheTaskAssign(task, first_task);
      if( ok )
      {
	gfs->soln_changed = true;
	if( gfs->task_set != NULL )
	  KheTaskSetAddTask(gfs->task_set, task);
      }
    }
    if( DEBUG3 )
    {
      fprintf(stderr, ", %s", ok ? "" : "not ");
      KheTaskDebug(task, 1, -1, stderr);
    }
  }
  if( DEBUG3 )
    fprintf(stderr, "}\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetGroupAndRemovePlateau(KHE_DOMAIN_TASK_SET dts,      */
/*    int start_pos, int end_pos, KHE_GROUP_BY_FRAME_SOLVER gfs)             */
/*                                                                           */
/*  There is a plateau at start_pos .. end_pos.  Group it and remove it.     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDomainTaskSetGroupAndRemovePlateau(KHE_DOMAIN_TASK_SET dts,
  int start_pos, int end_pos, KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  int i, len;  KHE_TASK_SET ts;

  ** optionally group the plateau **
  len = end_pos - start_pos + 1;
  if( len == 1 || len > dts->max_limit )
  {
    ** too small or oo large to group **
  }
  else if( len < 2 * dts->min_limit )
  {
    ** can't break in two, so group it all **
    KheDomainTaskSetGroup(dts, start_pos, end_pos, gfs);
  }
  else if( len == 2 * dts->min_limit )
  {
    ** two equal groups is the only option here **
    ** given that they can both be assigned to the same resource **
    KheDomainTaskSetGroup(dts, start_pos, start_pos + dts->min_limit - 1, gfs);
    KheDomainTaskSetGroup(dts, start_pos + dts->min_limit, end_pos, gfs);
  }
  else
  {
    ** probably no single solution, so leave it ungrouped **
  }

  ** remove the plateau **
  for( i = start_pos;  i <= end_pos;  i++ )
  {
    ts = HaArray(dts->task_sets, i);
    HnAssert(KheTaskSetTaskCount(ts) > 0,
      "KheDomainTaskSetGroupAndRemovePlateau internal error");
    KheTaskSetDropFromEnd(ts, 1);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetPlateauDebug(KHE_DOMAIN_TASK_SET dts,               */
/*    int start_pos, int end_pos, int verbosity, int indent, FILE *fp)       */
/*                                                                           */
/*  Debug print of a plateau of dts at the given verbosity.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDomainTaskSetPlateauDebug(KHE_DOMAIN_TASK_SET dts,
  int start_pos, int end_pos, int verbosity, int indent, FILE *fp)
{
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT laic; KHE_TIME t1, t2;
  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  laic = KheLimitActiveIntervalsMonitorConstraint(dts->monitor);
  tg = KheLimitActiveIntervalsMonitorTimeGroup(dts->monitor, start_pos, &po);
  t1 = KheTimeGroupTime(tg, 0);
  tg = KheLimitActiveIntervalsMonitorTimeGroup(dts->monitor, end_pos, &po);
  t2 = KheTimeGroupTime(tg, 0);
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "Plateau(%s at %s .. %s)",
    KheConstraintId((KHE_CONSTRAINT) laic), KheTimeId(t1), KheTimeId(t2));
  if( indent > 0 )
    fprintf(fp, "\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetDebug(KHE_DOMAIN_TASK_SET dts, int verbosity,       */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dts onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDomainTaskSetDebug(KHE_DOMAIN_TASK_SET dts, int verbosity,
  int indent, FILE *fp)
{
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT laic; KHE_TIME t; KHE_TASK_SET ts;
  KHE_TIME_GROUP tg;  KHE_POLARITY po;  int i, tg_count;
  fprintf(fp, "%*s[ DomainTaskSet(", indent, "");
  laic = KheLimitActiveIntervalsMonitorConstraint(dts->monitor);
  fprintf(fp, "%s, %d-%d)\n", KheConstraintId((KHE_CONSTRAINT) laic),
    dts->min_limit, dts->max_limit);
  KheResourceGroupDebug(dts->domain, verbosity, indent + 2, fp);
  tg_count = KheLimitActiveIntervalsMonitorTimeGroupCount(dts->monitor);
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(dts->monitor, i, &po);
    HnAssert(KheTimeGroupTimeCount(tg) == 1 && po == KHE_POSITIVE,
      "KheDomainTaskSetDebug internal error 1");
    t = KheTimeGroupTime(tg, 0);
    ts = HaArray(dts->task_sets, i);
    HnAssert(ts != NULL, "KheDomainTaskSetDebug internal error 2");
    fprintf(fp, "%*s  %s (%d): ", indent, "", KheTimeId(t),
      KheTaskSetTaskCount(ts));
    KheTaskSetDebug(ts, verbosity, 0, fp);
  }
  fprintf(fp, "%*s]\n", indent, "");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheContainsDomainTaskSet(KHE_GROUP_BY_FRAME_SOLVER gfs,             */
/*    KHE_RESOURCE_GROUP rg, KHE_DOMAIN_TASK_SET *res)                       */
/*                                                                           */
/*  If gfs contains a domain task set with domain rg, set *res to that       */
/*  domain task set and return true.  Otherwise return false.                */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheContainsDomainTaskSet(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_RESOURCE_GROUP rg, KHE_DOMAIN_TASK_SET *res)
{
  KHE_DOMAIN_TASK_SET dts;  int i;
  HaArrayForEach(gfs->domain_task_sets, dts, i)
    if( KheResourceGroupEqual(dts->domain, rg) )
      return *res = dts, true;
  return *res = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheMakeDomainTaskSets(KHE_GROUP_BY_FRAME_SOLVER gfs,                */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit, int max_limit) */
/*                                                                           */
/*  Make domain task sets for laim.                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheMakeDomainTaskSets(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit, int max_limit)
{
  KHE_TIME t;  KHE_TIME_GROUP tg;  int i, j, k, tg_count;  KHE_POLARITY po;
  KHE_MEET meet;  KHE_TASK task;  KHE_DOMAIN_TASK_SET dts;
  KHE_RESOURCE_TYPE rt;

  ** boilerplate **
  HaArrayClear(gfs->domain_task_sets);
  rt = gs->resource_type;

  ** find and add each task of the appropriate type running under laim **
  tg_count = KheLimitActiveIntervalsMonitorTimeGroupCount(laim);
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
    HnAssert(KheTimeGroupTimeCount(tg) == 1 && po == KHE_POSITIVE,
      "KheMakeDomainTaskSets internal error 1");
    t = KheTimeGroupTime(tg, 0);
    for( j = 0;  j < KheEventTimetableMonitorTimeMeetCount(gfs->etm, t);  j++ )
    {
      meet = KheEventTimetableMonitorTimeMeet(gfs->etm, t, j);
      if( KheMeetDuration(meet) == 1 )
	for( k = 0;  k < KheMeetTaskCount(meet);  k++ )
	{
	  task = KheTaskProperRoot(KheMeetTask(meet, k));
	  if( KheTaskResourceType(task) == rt &&
	      KheTaskNonAssignmentHasCost(task) )
	  {
	    ** add task to the appropriate domain task set **
            if( !KheContainsDomainTaskSet(gfs, KheTaskDomain(task), &dts) )
	    {
	      dts = KheDomainTaskSetMake(gfs, laim, KheTaskDomain(task),
		min_limit, max_limit);
	      HaArrayAddLast(gfs->domain_task_sets, dts);
	    }
	    KheDomainTaskSetAddTask(dts, task, i);
	  }
	}
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "framed monitors"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_FRAMED_MONITOR KheFramedMonitorMake(KHE_MONITOR m,                   */
/*    int first_index, int last_index)                                       */
/*                                                                           */
/*  Make a new framed monitor object with these attributes.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_FRAMED_MONITOR KheFramedMonitorMake(KHE_MONITOR m,
  int first_index, int last_index, HA_ARENA a)
{
  KHE_FRAMED_MONITOR res;
  HaMake(res, a);
  res->monitor = m;
  res->first_index = first_index;
  res->last_index = last_index;
  res->elim_lim = -1;
  HaArrayInit(res->elim_intersects, a);
  res->elim_supply = res->elim_demand = -1;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheFramedMonitorCmp(const void *t1, const void *t2)                  */
/*                                                                           */
/*  Untyped comparison function for sorting an array of framed monitor       */
/*  objects by increasing (first_index, last_index).                         */
/*                                                                           */
/*****************************************************************************/

/* ***
int KheFramedMonitorCmp(const void *t1, const void *t2)
{
  KHE_FRAMED_MONITOR fm1 = * (KHE_FRAMED_MONITOR *) t1;
  KHE_FRAMED_MONITOR fm2 = * (KHE_FRAMED_MONITOR *) t2;
  int cmp;
  cmp = fm1->first_index - fm2->first_index;
  if( cmp != 0 ) return cmp;
  return fm1->last_index - fm2->last_index;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitBusyTimesConstraintSingleResourceType(                      */
/*    KHE_LIMIT_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE *rt)              */
/*                                                                           */
/*  If the resources that c applies to have a single resource type, set      */
/*  *rt to that resource type and return true.  Otherwise return false.      */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheLimitBusyTimesConstraintSingleResourceType(
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE *rt)
{
  int i;  KHE_RESOURCE_GROUP rg;  KHE_RESOURCE r;
  *rt = NULL;
  for( i = 0;  i < KheLimitBusyTimesConstraintResourceGroupCount(c);  i++ )
  {
    rg = KheLimitBusyTimesConstraintResourceGroup(c, i);
    if( *rt == NULL )
      *rt = KheResourceGroupResourceType(rg);
    else if( *rt != KheResourceGroupResourceType(rg) )
      return *rt = NULL, false;
  }
  for( i = 0;  i < KheLimitBusyTimesConstraintResourceCount(c);  i++ )
  {
    r = KheLimitBusyTimesConstraintResource(c, i);
    if( *rt == NULL )
      *rt = KheResourceResourceType(r);
    else if( *rt != KheResourceResourceType(r) )
      return *rt = NULL, false;
  }
  return *rt != NULL;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheUseLimitBusyTimesMonitor(KHE_LIMIT_BUSY_TIMES_MONITOR lbtm,      */
/*    KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res)                       */
/*                                                                           */
/*  If lbtm is relevant, return true and set *res to its framed monitor      */
/*  object.  Otherwise return false.                                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheUseLimitBusyTimesMonitor(KHE_LIMIT_BUSY_TIMES_MONITOR lbtm,
  KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res, HA_ARENA a)
{
  int i, offset, c_count;  KHE_TIME_GROUP tg;  KHE_RESOURCE_TYPE rt;
  KHE_TIME t, earliest_t, latest_t;  KHE_LIMIT_BUSY_TIMES_CONSTRAINT lbtc;

  ** if monitor's constraint does not monitor all resources, return false **
  lbtc = KheLimitBusyTimesMonitorConstraint(lbtm);
  if( !KheLimitBusyTimesConstraintSingleResourceType(lbtc, &rt) )
    return *res = NULL, false;
  c_count = KheLimitBusyTimesConstraintResourceCount(lbtc);
  if( c_count < 0.8 * KheResourceTypeResourceCount(rt) )
    return *res = NULL, false;
  ** ***
  if( KheLimitBusyTimesConstraintResourceGroupCount(lbtc) != 1 )
    return *res = NULL, false;
  rg = KheLimitBusyTimesConstraintResourceGroup(lbtc, 0);
  if( KheResourceGroupResourceCount(rg) !=
      KheResourceTypeResourceCount(KheResourceGroupResourceType(rg)) )
    return *res = NULL, false;
  *** **

  ** find the earliest and latest times; if none, return false **
  earliest_t = latest_t = NULL;
  offset = KheLimitBusyTimesMonitorOffset(lbtm);
  for( i = 0;  i < KheLimitBusyTimesConstraintTimeGroupCount(lbtc);  i++ )
  {
    tg = KheLimitBusyTimesConstraintTimeGroup(lbtc, i, offset);
    if( KheTimeGroupTimeCount(tg) > 0 )
    {
      t = KheTimeGroupTime(tg, 0);
      if( earliest_t == NULL || KheTimeIndex(t) < KheTimeIndex(earliest_t) )
        earliest_t = t;
      t = KheTimeGroupTime(tg, KheTimeGroupTimeCount(tg) - 1);
      if( latest_t == NULL || KheTimeIndex(t) > KheTimeIndex(latest_t) )
        latest_t = t;
    }
  }
  if( earliest_t == NULL )
    return *res = NULL, false;

  ** all good, return a framed monitor object **
  return *res = KheFramedMonitorMake((KHE_MONITOR) lbtm,
    KheFrameTimeIndex(common_frame, earliest_t),
    KheFrameTimeIndex(common_frame, latest_t), a), true;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheUseMonitor(KHE_MONITOR m, KHE_FRAME common_frame,                */
/*    KHE_FRAMED_MONITOR *res)                                               */
/*                                                                           */
/*  If m is relevant, return true and set *res to its framed monitor         */
/*  object.  Otherwise return false.                                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheUseMonitor(KHE_MONITOR m, KHE_FRAME common_frame,
  KHE_FRAMED_MONITOR *res, HA_ARENA a)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      return KheUseClusterBusyTimesMonitor((KHE_CLUSTER_BUSY_TIMES_MONITOR) m,
        common_frame, res, a);

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      return KheUseLimitBusyTimesMonitor((KHE_LIMIT_BUSY_TIMES_MONITOR) m,
        common_frame, res, a);

    default:

      ** ignore other kinds of monitors **
      return *res = NULL, false;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheFramedMonitorDebug(KHE_FRAMED_MONITOR fm, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of fm onto fp with the given indent.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheFramedMonitorDebug(KHE_FRAMED_MONITOR fm, int indent, FILE *fp)
{
  int val, i;  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm;  KHE_POLARITY po;
  KHE_TIME_GROUP tg;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Framed monitor %d-%d\n", indent, "",
      fm->first_index, fm->last_index);
    KheMonitorDebug(fm->monitor, 2, indent + 2, fp);
    if( fm->elim_lim >= 0 )
    {
      cbtm = (KHE_CLUSTER_BUSY_TIMES_MONITOR) fm->monitor;
      HaArrayForEach(fm->elim_intersects, val, i)
      {
	fprintf(fp, "%*s  %d time groups of frame intersect with ",
	  indent, "", val);
	tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
	KheTimeGroupDebug(tg, 1, 0, fp);
      }
      fprintf(fp, "%*s  supply %d, demand %d\n", indent, "",
        fm->elim_supply, fm->elim_demand);
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "group by frame solvers"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_GROUP_BY_FRAME_SOLVER KheGroupByFrameSolverMake(                     */
/*    KHE_TASKING tasking, int max_days, bool resource_invariant,            */
/*    KHE_FRAME common_frame, KHE_EVENT_TIMETABLE_MONITOR etm)               */
/*                                                                           */
/*  Make a new group by frame solver object.                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_GROUP_BY_FRAME_SOLVER KheGroupByFrameSolverMake(
  KHE_TASKING tasking, int max_days, bool resource_invariant,
  KHE_FRAME common_frame, KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_TASK_SET ts, HA_ARENA a)
{
  KHE_GROUP_BY_FRAME_SOLVER res;  int count;  KHE_INSTANCE ins;

  ** the solve as a whole **
  HaMake(res, a);
  res->arena = a;
  res->tasking = tasking;
  res->soln = KheTaskingSoln(tasking);
  res->resource_invariant = resource_invariant;
  res->common_frame = common_frame;
  res->etm = etm;
  res->resource = NULL;
  res->soln_changed = false;
  res->task_set = ts;

  ** phase 0 **
  count = KheFrameTimeGroupCount(common_frame);
  HaArrayInit(res->provisionally_linked_to_prev, a);
  HaArrayFill(res->provisionally_linked_to_prev, count, false);
  HaArrayInit(res->linked_to_prev, a);
  HaArrayFill(res->linked_to_prev, count, false);

  ** phase 1 **
  res->max_days = max_days;
  HaArrayInit(res->framed_monitors, a);
  HaArrayInit(res->start_indexes, a);
  res->group_monitor = NULL;
  ins = KheSolnInst ance(res->soln);
  res->successful_times = KheTimeSetMake(ins, a);
  res->last_index = 0;
  res->success_count = 0;
  res->success_ts = KheTimeSetMake(ins, a);
  res->curr_ts = KheTimeSetMake(ins, a);

  ** phase 2 **
  HaArrayInit(res->time_sets, a);
  HaArrayInit(res->time_sets_by_time, a);
  HaArrayInit(res->time_sets_by_monitor, a);
  HaArrayInit(res->suitable_monitors, a);
  HaArrayInit(res->success_seq, a);
  HaArrayInit(res->curr_seq, a);
  HaArrayInit(res->task_sets, a);
  HaArrayInit(res->counts, a);
  HaArrayInit(res->domain_task_sets, a);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheTaskCmp(const void *t1, const void *t2)                           */
/*                                                                           */
/*  Comparison function for sorting tasks so that they can be uniqueified.   */
/*                                                                           */
/*****************************************************************************/

/* ***
static int KheTaskCmp(const void *t1, const void *t2)
{
  KHE_TASK task1 = * (KHE_TASK *) t1;
  KHE_TASK task2 = * (KHE_TASK *) t2;
  return KheTaskSolnIndex(task1) - KheTaskSolnIndex(task2);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_TASK_SET KheFindAllTasksAtTime(KHE_GROUP_BY_FRAME_SOLVER gfs,        */
/*    KHE_TIME time)                                                         */
/*                                                                           */
/*  Return a new task set containing all the tasks of duration 1 running     */
/*  at time, or rather, their proper roots, uniqueified.  Only tasks with    */
/*  the appropriate type, and only tasks which would incur a cost if not     */
/*  assigned, are included.                                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_TASK_SET KheFindAllTasksAtTime(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_TIME time)
{
  KHE_TASK_SET res;  KHE_MEET meet;  KHE_TASK task;  int i, j;
  KHE_RESOURCE_TYPE rt;
  rt = KheResourceResourceType(gfs->resource);
  res = KheTaskSetMake(gfs->soln);
  for( i = 0; i < KheEventTimetableMonitorTimeMeetCount(gfs->etm, time); i++ )
  {
    meet = KheEventTimetableMonitorTimeMeet(gfs->etm, time, i);
    if( KheMeetDuration(meet) == 1 )
      for( j = 0;  j < KheMeetTaskCount(meet);  j++ )
      {
	task = KheTaskProperRoot(KheMeetTask(meet, j));
	if( KheTaskResourceType(task)==rt && KheTaskNonAssignmentHasCost(task) )
	  KheTaskSetAddTask(res, task);
      }
  }
  KheTaskSetSortUnique(res, &KheTaskCmp);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_TASK_SET KheFindTasksAtTime(KHE_RCG_SOLVER gs, KHE_TIME time)*/
/*                                                                           */
/*  Return a new task set containing all the tasks of duration 1 running     */
/*  at time, or rather, their proper roots, uniqueified.  Only tasks with    */
/*  the appropriate type, and only tasks which would incur a cost if not     */
/*  assigned, are included.                                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_TASK_SET KheFindTasksAtTime(KHE_RCG_SOLVER gs, KHE_TIME time)
{
  KHE_TASK_SET res;  KHE_MEET meet;  KHE_TASK task;  int i, j;
  KHE_RESOURCE_TYPE rt;
  rt = KheResourceResourceType(gs->resource);
  res = KheTaskSetMake(gs->soln);
  for( i = 0; i < KheEventTimetableMonitorTimeMeetCount(gs->etm, time); i++ )
  {
    meet = KheEventTimetableMonitorTimeMeet(gs->etm, time, i);
    if( KheMeetDuration(meet) == 1 )
      for( j = 0;  j < KheMeetTaskCount(meet);  j++ )
      {
	task = KheTaskProperRoot(KheMeetTask(meet, j));
	if( KheTaskResourceType(task) == rt && KheTaskTotalDuration(task) == 1 
	    && KheTaskNonAssignmentHasCost(task) )
	  KheTaskSetAddTask(res, task);
      }
  }
  KheTaskSetSortUnique(res, &KheTaskCmp);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheFindAndAssignTaskAtTime(KHE_RCG_SOLVER gs,               */
/*    KHE_TIME time, KHE_TASK *task)                                         */
/*                                                                           */
/*  Find and assign an arbitrary unassigned task of duration 1 at this       */
/*  time, and return true if successful.                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheFindAndAssignTaskAtTime(KHE_RCG_SOLVER gs,
  KHE_TIME time, KHE_TASK *task)
{
  KHE_MEET meet;  KHE_TASK task2;  int i, j;
  for( i = 0; i < KheEventTimetableMonitorTimeMeetCount(gs->etm, time); i++ )
  {
    meet = KheEventTimetableMonitorTimeMeet(gs->etm, time, i);
    if( KheMeetDuration(meet) == 1 )
      for( j = 0;  j < KheMeetTaskCount(meet);  j++ )
      {
	task2 = KheTaskProperRoot(KheMeetTask(meet, j));
	if( KheTaskAssignResource(task2, gs->resource) )
	  return *task = task2, true;
      }
  }
  return *task = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheClusterBusyTimesConstraintSingleResourceType(                    */
/*    KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE *rt)            */
/*                                                                           */
/*  If the resources that c applies to have a single resource type, set      */
/*  *rt to that resource type and return true.  Otherwise return false.      */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheClusterBusyTimesConstraintSingleResourceType(
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE *rt)
{
  int i;  KHE_RESOURCE_GROUP rg;  KHE_RESOURCE r;
  *rt = NULL;
  for( i = 0;  i < KheClusterBusyTimesConstraintResourceGroupCount(c);  i++ )
  {
    rg = KheClusterBusyTimesConstraintResourceGroup(c, i);
    if( *rt == NULL )
      *rt = KheResourceGroupResourceType(rg);
    else if( *rt != KheResourceGroupResourceType(rg) )
      return *rt = NULL, false;
  }
  for( i = 0;  i < KheClusterBusyTimesConstraintResourceCount(c);  i++ )
  {
    r = KheClusterBusyTimesConstraintResource(c, i);
    if( *rt == NULL )
      *rt = KheResourceResourceType(r);
    else if( *rt != KheResourceResourceType(r) )
      return *rt = NULL, false;
  }
  return *rt != NULL;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheUseClusterBusyTimesMonitor(KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm,  */
/*    KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res)                       */
/*                                                                           */
/*  If cbtm is relevant, return true and set *res to its framed monitor      */
/*  object.  Otherwise return false.                                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheUseClusterBusyTimesMonitor(KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm,
  KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res, HA_ARENA a)
{
  int i, c_count;  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_RESOURCE_TYPE rt;
  KHE_TIME t, earliest_t, latest_t;  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;

  ** if monitor's constraint does not monitor most resources, return false **
  cbtc = KheClusterBusyTimesMonitorConstraint(cbtm);
  if( !KheClusterBusyTimesConstraintSingleResourceType(cbtc, &rt) )
    return *res = NULL, false;
  c_count = KheClusterBusyTimesConstraintResourceCount(cbtc);
  if( c_count < 0.8 * KheResourceTypeResourceCount(rt) )
    return *res = NULL, false;
  ** ***
  if( KheClusterBusyTimesConstraintResourceGroupCount(cbtc) != 1 )
    return *res = NULL, false;
  if( KheResourceGroupResourceCount(rg) !=
      KheResourceTypeResourceCount(KheResourceGroupResourceType(rg)) )
    return *res = NULL, false;
  *** **

  ** find the earliest and latest times; if none, return false **
  earliest_t = latest_t = NULL;
  for( i = 0;  i < KheClusterBusyTimesMonitorTimeGroupCount(cbtm);  i++ )
  {
    tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
    if( KheTimeGroupTimeCount(tg) > 0 )
    {
      t = KheTimeGroupTime(tg, 0);
      if( earliest_t == NULL || KheTimeIndex(t) < KheTimeIndex(earliest_t) )
        earliest_t = t;
      t = KheTimeGroupTime(tg, KheTimeGroupTimeCount(tg) - 1);
      if( latest_t == NULL || KheTimeIndex(t) > KheTimeIndex(latest_t) )
        latest_t = t;
    }
  }
  if( earliest_t == NULL )
    return *res = NULL, false;

  ** all good, make and return a framed monitor object **
  *res = KheFramedMonitorMake((KHE_MONITOR) cbtm,
    KheFrameTimeIndex(common_frame, earliest_t),
    KheFrameTimeIndex(common_frame, latest_t), a);
    
  return true;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDemandAtClusterMonitor(KHE_EVENT_TIMETABLE_MONITOR etm,           */
/*    KHE_RESOURCE_TYPE rt, KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm)             */
/*                                                                           */
/*  Return the total demand for resources of type rt in the time groups      */
/*  of cluster busy times monitor cbtm.                                      */
/*                                                                           */
/*****************************************************************************/

/* ***
static int KheDemandAtClusterMonitor(KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_RESOURCE_TYPE rt, KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm)
{
  int i, res;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  res = 0;
  for( i = 0;  i < KheClusterBusyTimesMonitorTimeGroupCount(cbtm);  i++ )
  {
    tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
    res += KheDemandAtTimeGroup(etm, rt, tg);
  }
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheEliminateCombinationsForMonitor(KHE_GROUP_BY_FRAME_SOLVER gfs,   */
/*    KHE_FRAMED_MONITOR fm, int lim)                                        */
/*                                                                           */
/*  Eliminate combinations based on supply and demand within the times of    */
/*  fm, which holds a suitable cluster busy times monitor with limit lim.    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheEliminateCombinationsForMonitor(KHE_RCG_SOLVER gs,
  KHE_FRAMED_MONITOR fm, int lim)
{
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm;  KHE_TIME_GROUP m_tg, f_tg, prev_f_tg;
  int i, j, count, sum;  KHE_POLARITY po;
  bool f_tg_overlaps, prev_f_tg_overlaps;  KHE_TIME_GROUP_INFO tgi;
  if( DEBUG6 )
  {
    fprintf(stderr, "[ KheEliminateCombinationsForMonitor(gs, fm, %d), cbtm \n",
      lim);
    KheMonitorDebug(fm->monitor, 3, 2, stderr);
  }

  ** for each time group of fm, find the number of frame time groups it **
  ** intersects with.  Also record in ec_time_group_info those time groups **
  ** which will be linked with previous time groups if demand allows **
  HaArrayClear(gs->ec_time_group_info);
  fm->elim_lim = lim;
  cbtm = (KHE_CLUSTER_BUSY_TIMES_MONITOR) fm->monitor;
  for( i = 0;  i < KheClusterBusyTimesMonitorTimeGroupCount(cbtm);  i++ )
  {
    m_tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
    count = 0;
    prev_f_tg = NULL;
    prev_f_tg_overlaps = false;
    for( j = fm->first_index;  j <= fm->last_index;  j++ )
    {
      f_tg = KheFrameTimeGroup(gs->common_frame, j);
      f_tg_overlaps = !KheTimeGroupDisjoint(m_tg, f_tg);
      if( f_tg_overlaps )
	count++;
      if( f_tg_overlaps && prev_f_tg_overlaps )
      {
	** if we later find that D(T) >= S(T), then if prev_f_tg is **
	** busy, then f_tg must be busy as well, because they both  **
	** overlap the same time group of the monitor **
	if( DEBUG6 )
	{
	  fprintf(stderr, "  provisionally linked time groups ");
	  KheTimeGroupDebug(prev_f_tg, 1, -1, stderr);
	  fprintf(stderr, " and ");
	  KheTimeGroupDebug(f_tg, 1, -1, stderr);
	  fprintf(stderr, "\n");
	}

	** record the linkage provisionally **
	tgi = HaArray(gs->time_group_info, j);
	HaArrayAddLast(gs->ec_time_group_info, tgi);
      }
      prev_f_tg = f_tg;
      prev_f_tg_overlaps = f_tg_overlaps;
    }
    HaArrayAddLast(fm->elim_intersects, count);
  }

  ** get the supply:  the largest lim intersect counts, times resources **
  HaArraySort(fm->elim_intersects, &IntCmpIncreasing);
  count = HaArrayCount(fm->elim_intersects);
  sum = 0;
  for( i = 1;  i <= lim;  i++ )
    sum += HaArray(fm->elim_intersects, count - i);
  fm->elim_supply = sum * KheResourceTypeResourceCount(gs->resource_type);

  ** get the demand **
  fm->elim_demand = KheDemandAtClusterMonitor(gs->etm, gs->resource_type, cbtm);

  ** if demand >= supply, provisional linkages become definite **
  if( fm->elim_demand >= fm->elim_supply )
    HaArrayForEach(gs->ec_time_group_info, tgi, i)
      tgi->linked_to_prev = true;

  if( DEBUG6 )
  {
    KheFramedMonitorDebug(fm, 2, stderr);
    fprintf(stderr, "] KheEliminateCombinationsForMonitor returning\n");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameGroupTasks(KHE_GROUP_BY_FRAME_SOLVER gfs,            */
/*     KHE_TIME_SET ts)                                                      */
/*                                                                           */
/*  Group the tasks identified by ts, and set gfs->soln_changed if some      */
/*  actual grouping occurs.                                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheGroupByFrameGroupTasks(KHE_GROUP_BY_FRAME_SOLVER gfs,
   KHE_TIME_SET ts)
{
  KHE_TASK_SET task_set;  KHE_TIME t;
  KHE_TASK first_task, task;  int i, j, count, min_count;  bool ok;
  if( KheTimeSetTimeCount(ts) >= 2 )
  {
    ** get one task set for each time; also calculate the smallest size **
    HaArrayClear(gfs->task_sets);
    min_count = INT_MAX;
    ** HaArrayForEach(ts->times, t, i) **
    for( i = 0;  i < KheTimeSetTimeCount(ts);  i++ )
    {
      t = KheTimeSetTime(ts, i);
      task_set = KheFindAllTasksAtTime(gfs, t);
      HaArrayAddLast(gfs->task_sets, task_set);
      count = KheTaskSetTaskCount(task_set);
      if( count < min_count )
	min_count = count;
    }

    ** carry out the groupings; NB none of the tasks are currently assigned **
    for( i = 0;  i < min_count;  i++ )
    {
      ** group together the i'th tasks from each task set **
      task_set = HaArray(gfs->task_sets, 0);
      first_task = KheTaskSetTask(task_set, i);
      if( DEBUG3 )
      {
	fprintf(stderr, "    KheGroupByFrame grouping tasks {");
	KheTaskDebug(first_task, 1, -1, stderr);
      }
      for( j = 1;  j < HaArrayCount(gfs->task_sets);  j++ )
      {
	task_set = HaArray(gfs->task_sets, j);
	task = KheTaskSetTask(task_set, i);
	if( task == first_task )
	  ok = true;
	else
	{
	  ok = KheTaskAssign(task, first_task);
	  if( ok )
	  {
	    gfs->soln_changed = true;
	    if( gfs->task_set != NULL )
	      KheTaskSetAddTask(gfs->task_set, task);
	  }
	}
	if( DEBUG3 )
	{
	  fprintf(stderr, ", %s", ok ? "" : "not ");
	  KheTaskDebug(task, 1, -1, stderr);
	}
      }
      if( DEBUG3 )
	fprintf(stderr, "}\n");
    }

    ** free the task sets **
    HaArrayForEach(gfs->task_sets, task_set, i)
      KheTaskSetDelete(task_set);
    HaArrayClear(gfs->task_sets);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameSolverDebug(KHE_GROUP_BY_FRAME_SOLVER gfs,           */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Produce a debug print of gfs on fp with the given indent.                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheGroupByFrameSolverDebug(KHE_GROUP_BY_FRAME_SOLVER gfs,
  int indent, FILE *fp)
{
  KHE_FRAMED_MONITOR fm;  int i, pos;  KHE_TIME_GROUP tg;  bool val;
  fprintf(fp, "%*s[ GroupByFrameSolver(%s, r %s, max_days %d, rs_invt %s)\n", 
    indent, "", KheInstanceId(KheSoln Instance(gfs->soln)),
    KheResourceId(gfs->resource), gfs->max_days,
    gfs->resource_invariant ? "true" : "false");
  fprintf(fp, "  [ common frame (# indicates linked to previous):\n");
  for( i = 0;  i < KheFrameTimeGroupCount(gfs->common_frame);  i++ )
  {
    tg = KheFrameTimeGroup(gfs->common_frame, i);
    val = HaArray(gfs->linked_to_prev, i);
    fprintf(fp, "    %c ", val ? '#' : ' ');
    KheTimeGroupDebug(tg, 1, 0, fp);
  }
  fprintf(fp, "  ]\n");
  HaArrayForEach(gfs->framed_monitors, fm, i)
  {
    fprintf(fp, "%*s%2d: ", indent, "", i);
    KheFramedMonitorDebug(fm, 0, fp);
  }
  fprintf(fp, "%*s[", indent + 2, "");
  HaArrayForEach(gfs->start_indexes, pos, i)
    fprintf(fp, "%s%d", i > 0 ? ", " : "", pos);
  fprintf(fp, "]\n");
  fprintf(fp, "%*s]\n", indent, "");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheFramedMonitorsBegin(KHE_RCG_SOLVER gs)                   */
/*                                                                           */
/*  Build framed monitors for gs.                                            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheFramedMonitorsBegin(KHE_RCG_SOLVER gs)
{
  KHE_FRAMED_MONITOR fm;  int i, pos, si;  KHE_MONITOR m;

  ** build and sort the framed_monitors array **
  HnAssert(HaArrayCount(gs->framed_monitors) == 0,
    "KheFramedMonitorsBegin internal error 1");
  for( i = 0;  i < KheSolnResourceMonitorCount(gs->soln, gs->resource);  i++ )
  {
    m = KheSolnResourceMonitor(gs->soln, gs->resource, i);
    if( KheUseMonitor(m, gs->common_frame, &fm, gs->arena) )
      HaArrayAddLast(gs->framed_monitors, fm);
  }
  HaArraySort(gs->framed_monitors, &KheFramedMonitorCmp);

  ** build the start_indexes array **
  HnAssert(HaArrayCount(gs->start_indexes) == 0,
    "KheFramedMonitorsBegin internal error 2");
  pos = 0;
  for( si = 0;  pos < HaArrayCount(gs->framed_monitors);  si++ )
  {
    ** add the index for si **
    fm = HaArray(gs->framed_monitors, pos);
    while( fm->first_index < si )
    {
      pos++;
      if( pos >= HaArrayCount(gs->framed_monitors) )
	break;
      fm = HaArray(gs->framed_monitors, pos);
    }
    HaArrayAddLast(gs->start_indexes, pos);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheFramedMonitorsEnd(KHE_RCG_SOLVER gs)                     */
/*                                                                           */
/*  Clear out the framed monitors and their associated start_indexes.        */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheFramedMonitorsEnd(KHE_RCG_SOLVER gs)
{
  HaArrayClear(gs->framed_monitors);
  HaArrayClear(gs->start_indexes);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_GROUP_MONITOR KheGroupMonitorForInterval(KHE_RCG_SOLVER gs,  */
/*    int first_index, int last_index)                                       */
/*                                                                           */
/*  Make a group monitor out of all the framed monitors that monitor time    */
/*  groups lying entirely within first_index .. last_index of the common     */
/*  frame.                                                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_GROUP_MONITOR KheGroupMonitorForInterval(KHE_RCG_SOLVER gs,
  int first_index, int last_index)
{
  KHE_GROUP_MONITOR res;  int i, si;  KHE_FRAMED_MONITOR fm;
  res = KheGroupMonitorMake(gs->soln, 1, "GroupByFrame");
  si = (first_index < HaArrayCount(gs->start_indexes) ?
    HaArray(gs->start_indexes, first_index) :
    HaArrayCount(gs->framed_monitors));
  for( i = si;  i < HaArrayCount(gs->framed_monitors);  i++ )
  {
    ** fm starts at or after first_index **
    fm = HaArray(gs->framed_monitors, i);
    HnAssert(fm->first_index >= first_index,
      "KheGroupMonitorForInterval internal error");
    if( fm->last_index <= last_index )
    {
      ** fm ends at or before last_index **
      KheGroupMonitorAddChildMonitor(res, fm->monitor);
      if( DEBUG2 )
      {
	fprintf(stderr, "    adding ");
	KheFramedMonitorDebug(fm, 0, stderr);
      }
    }

    ** if fm->first_index > last_index, there is no point in going further **
    if( fm->first_index > last_index )
      break;
  }
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFramePhase1ForInterval(KHE_GROUP_BY_FRAME_SOLVER gfs,     */
/*    int first_index, int last_index)                                       */
/*                                                                           */
/*  Solve gfs for the interval first_index .. last_index, and return         */
/*  true if there was exactly one success, with gfs->success_ts holding      */
/*  the successful set of times.                                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheGroupByFramePhase1ForInterval(KHE_GROUP_BY_FRAME_SOLVER gfs,
  int first_index, int last_index)
{
  int i, si;  bool success;  KHE_FRAMED_MONITOR fm;  KHE_TIME_GROUP tg;
  KHE_TIME t;  KHE_TASK task;  KHE_TIME_SET ts;  KHE_INSTANCE ins;
  if( DEBUG2 )
    fprintf(stderr,
      "  [ KheGroupByFrameSolverFindAsst(gfs, first %d, last %d)\n",
      first_index, last_index);
  ins = KheSolnIn stance(gfs->soln);

  ** gather relevant monitors into a group monitor **
  gfs->group_monitor = KheGroupMonitorMake(gfs->soln, 1, "GroupByFrame");
  si = (first_index < HaArrayCount(gfs->start_indexes) ?
    HaArray(gfs->start_indexes, first_index) :
    HaArrayCount(gfs->framed_monitors));
  for( i = si;  i < HaArrayCount(gfs->framed_monitors);  i++ )
  {
    ** fm starts at or after first_index **
    fm = HaArray(gfs->framed_monitors, i);
    HnAssert(fm->first_index >= first_index,
      "KheGroupByFrameSolverFindAsst internal error");
    if( fm->last_index <= last_index )
    {
      KheGroupMonitorAddChildMonitor(gfs->group_monitor, fm->monitor);
      if( DEBUG2 )
      {
	fprintf(stderr, "    adding ");
	KheFramedMonitorDebug(fm, 0, stderr);
      }
    }

    ** if fm->first_index > last_index, there is no point in going further **
    if( fm->first_index > last_index )
      break;
  }

  ** try one solve for each time in the first time group, but then **
  ** start all over again and have another go if any were successful **
  tg = KheFrameTimeGroup(gfs->common_frame, first_index);
  KheTimeSetClear(gfs->successful_times);
  do
  {
    success = false;
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      ** initialize the solve fields other than gm **
      gfs->last_index = last_index;
      gfs->success_count = 0;
      KheTimeSetClear(gfs->success_ts);
      KheTimeSetClear(gfs->curr_ts);

      ** solve for t, unless already successful **
      t = KheTimeGroupTime(tg, i);
      if( !KheTimeSetContainsTime(gfs->successful_times, t) )
      {
	if( KheFindAndAssignTaskAtTime(gfs, t, &task) )
	{
	  KheTimeSetAddTime(gfs->curr_ts, t);
	  KheCombinatorialGroupingAssign(gfs, first_index + 1, true);
	  KheTimeSetDeleteLastTime(gfs->curr_ts);
	  ** HaArrayDeleteLast(gfs->curr_ts->times); **
	  KheTaskUnAssign(task);
	}

	** if exactly one success, we have succeeded **
	if( gfs->success_count == 1 )
	{
	  ** have a single successful assignment, so group tasks **
	  if( DEBUG2 )
	  {
	    fprintf(stderr, "  KheGroupByFrameSolverFindAsst found group: ");
	    KheTimeSetDebug(gfs->success_ts, 2, 0, stderr);
	  }
	  ts = KheTimeSetMake(ins, gfs->arena);
	  KheTimeSetUnion(ts, gfs->success_ts);
	  ** KheTimeSetCopy(ts, gfs->success_ts); **
	  HaArrayAddLast(gfs->time_sets, ts);
	  KheGroupByFrameGroupTasks(gfs, ts);
          KheTimeSetAddTime(gfs->successful_times, t);
	  success = true;
	}
      }
    }
  } while( success );

  ** clear out for next (rt, f), and return **
  KheGroupMonitorDelete(gfs->group_monitor);
  gfs->group_monitor = NULL;
  if( DEBUG2 )
    fprintf(stderr, "  ] KheGroupByFrameSolverFindAsst returning\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void FindSeq(KHE_GROUP_BY_FRAME_SOLVER gfs, int min_limit,               */
/*    int max_limit, int remaining_len)                                      */
/*                                                                           */
/*  This is the recursive part of finding a unique sequence of               */
/*  sub-interval lengths, if there is one.  Each length must lie             */
/*  between min_limit and max_limit; the remaining wanted length             */
/*  is remaining_len.                                                        */
/*                                                                           */
/*****************************************************************************/

/* ***
static void FindSeq(KHE_GROUP_BY_FRAME_SOLVER gfs, int min_limit,
  int max_limit, int remaining_len)
{
  int first_len, len, i;
  if( ++(gfs->node_count) > 100 )
  {
    ** give up, search is taking too long **
    if( DEBUG5 )
      fprintf(stderr, "  curr_seq search taking too long\n");
    gfs->success_count = 2;
  }
  else if( remaining_len < 0 )
  {
    ** no solutions here **
  }
  else if( remaining_len == 0 )
  {
    ** curr_seq is a feasible soln; but no good if more than one **
    if( DEBUG5 )
    {
      fprintf(stderr, "  successful curr_seq:");
      HaArrayForEach(gfs->curr_seq, len, i)
	fprintf(stderr, " %d", len);
      fprintf(stderr, "\n");
    }
    gfs->success_count++;
    if( gfs->success_count >= 2 )
      return;

    ** if any of the lengths of curr_seq differ, there is more than one soln **
    first_len = HaArrayFirst(gfs->curr_seq);
    for( i = 1;  i < HaArrayCount(gfs->curr_seq);  i++ )
      if( HaArray(gfs->curr_seq, i) != first_len )
      {
	gfs->success_count = 2;
	return;
      }

    ** save curr_seq; at the moment it is the unique feasible solution **
    HaArrayClear(gfs->success_seq);
    HaArrayForEach(gfs->curr_seq, len, i)
      HaArrayAddLast(gfs->success_seq, len);
  }
  else ** remaining_len > 0 **
  {
    for( len = min_limit;  len <= max_limit && gfs->success_count < 2;  len++ )
    {
      HaArrayAddLast(gfs->curr_seq, len);
      FindSeq(gfs, min_limit, max_limit, remaining_len - len);
      HaArrayDeleteLast(gfs->curr_seq);
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheMergeOneGap(KHE_GROUP_BY_FRAME_SOLVER gfs,                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit,                */
/*    int max_limit, int gap_start, int gap_stop)                            */
/*                                                                           */
/*  Do merges in the gap from gap_start inclusive to gap_stop exclusive.     */
/*  Return true if successful.                                               */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheMergeOneGap(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit,
  int max_limit, int gap_start, int gap_stop)
{
  int i, j, len, start;  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_TIME_SET ts;
  KHE_INSTANCE ins;

  if( DEBUG8 )
  {
    fprintf(stderr, "  [ KheMergeOneGap(gfs, lim[%d..%d], gap[%d..%d]:\n",
      min_limit, max_limit, gap_start, gap_stop - 1);
    for( i = gap_start;  i < gap_stop;  i++ )
    {
      tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
      KheTimeGroupDebug(tg, 1, 6, stderr);
    }
  }

  ins = KheSolnInstance(gfs->soln);
  gfs->success_count = 0;
  HaArrayClear(gfs->success_seq);
  HaArrayClear(gfs->curr_seq);
  gfs->node_count = 0;
  FindSeq(gfs, min_limit, max_limit, gap_stop - gap_start);
  if( DEBUG8 )
  {
    fprintf(stderr, "      found %d solution%s", gfs->success_count,
      gfs->success_count != 1 ? "s" : "");
    if( gfs->success_count == 1 )
    {
      fprintf(stderr, ":");
      HaArrayForEach(gfs->success_seq, len, i)
	fprintf(stderr, " %d", len);
    }
    fprintf(stderr, "\n");
  }
  if( gfs->success_count == 1 )
  {
    ** merge the sets defined by the successful sequence **
    start = gap_start;
    ts = KheTimeSetMake(ins, gfs->arena);
    HaArrayForEach(gfs->success_seq, len, i)
    {
      KheTimeSetClear(ts);
      for( j = 0;  j < len;  j++ )
      {
	tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, start + j, &po);
	KheTimeSetAddTime(ts, KheTimeGroupTime(tg, 0));
      }
      if( DEBUG8 )
      {
	fprintf(stderr, "      merging time set ");
	KheTimeSetDebug(ts, 2, 0, stderr);
      }
      KheGroupByFrameGroupTasks(gfs, ts);
      start += len;
    }
  }

  if( DEBUG8 )
    fprintf(stderr, "  ] KheMergeOneGap returning %s\n",
      bool_show(gfs->success_count == 1));
  return gfs->success_count == 1;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KhePlateauMerge(KHE_GROUP_BY_FRAME_SOLVER gfs,                      */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit, int max_limit) */
/*                                                                           */
/*  Try the alternative method of merging within one gap, based on           */
/*  the number of tasks beginning and ending.                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KhePlateauMerge(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit, int max_limit)
{
  int i, start_pos, end_pos;  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT laic;
  KHE_DOMAIN_TASK_SET dts;

  ** boilerplate **
  if( DEBUG7 )
  {
    laic = KheLimitActiveIntervalsMonitorConstraint(laim);
    fprintf(stderr, "  [ KhePlateauMerge(gfs, %s, %d..%d)\n",
      KheConstraintId((KHE_CONSTRAINT) laic), min_limit, max_limit);
  }

  ** build the domain task sets **
  KheMakeDomainTaskSets(gfs, laim, min_limit, max_limit);

  ** for each domain task set, find and remove plateaus **
  HaArrayForEach(gfs->domain_task_sets, dts, i)
  {
    if( DEBUG7 )
      KheDomainTaskSetDebug(dts, 2, 4, stderr);
    while( KheDomainTaskSetFindPlateau(dts, &start_pos, &end_pos) )
    {
      if( DEBUG7 )
	KheDomainTaskSetPlateauDebug(dts, start_pos, end_pos, 2, 6, stderr);
      KheDomainTaskSetGroupAndRemovePlateau(dts, start_pos, end_pos, gfs);
    }
  }

  ** free the domain task sets **
  ** sti ll to do **
  if( DEBUG7 )
    fprintf(stderr, "  ] KhePlateauMerge returning\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDoProfileGroupingMonitors(KHE_GROUP_BY_FRAME_SOLVER gfs,         */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim1,                              */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim2)                              */
/*                                                                           */
/*  Carry out profile grouping based on laim1 and (optionally) laim2.        */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDoProfileGroupingMonitors(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim1,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim2)
{
  int i, j, min_limit, max_limit, tg_count, index;
  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_TIME t;
  KHE_TIME_SET ts, prev_ts, foll_ts;
  bool history_before, history_after;
  if( DEBUG9 )
  {
    fprintf(stderr, "[ KheDoProfileGroupingMonitors(gfs, laim1, %s)\n",
      laim2 != NULL ? "laim2" : "-");
    KheMonitorDebug((KHE_MONITOR) laim1, 1, 2, stderr);
    if( laim2 != NULL )
      KheMonitorDebug((KHE_MONITOR) laim2, 1, 2, stderr);
  }

  ** set up the time_sets_by_monitor array **
  if( DEBUG9 )
    fprintf(stderr, "    [ time_sets_by_monitor:\n");
  HaArrayClear(gfs->time_sets_by_monitor);
  tg_count = KheLimitActiveIntervalsMonitorTimeGroupCount(laim1);
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim1, i, &po);
    HnAssert(KheTimeGroupTimeCount(tg) == 1 && po == KHE_POSITIVE,
      "KheDoProfileGroupingMonitors internal error");
    t = KheTimeGroupTime(tg, 0);
    ts = HaArray(gfs->time_sets_by_time, KheTimeIndex(t));
    HaArrayAddLast(gfs->time_sets_by_monitor, ts);
    if( DEBUG9 )
    {
      fprintf(stderr, "        %d: ", i);
      KheTimeGroupDebug(tg, 1, -1, stderr);
      fprintf(stderr, " --> ");
      if( ts != NULL )
	KheTimeSetDebug(ts, 2, 0, stderr);
      else
	fprintf(stderr, "NULL\n");
    }
  }
  if( DEBUG9 )
    fprintf(stderr, "    ]\n");

  ** if time sets do not coordinate with time groups, exit early **
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim1, i, &po);
    ts = HaArray(gfs->time_sets_by_monitor, i);
    if( ts != NULL )
    {
      for( j = i + 1;  j < tg_count;  j++ )
      {
	if( HaArray(gfs->time_sets_by_monitor, j) != ts )
	  break;
      }
      if( j - i != KheTimeSetTimeCount(ts) )
      {
	if( DEBUG9 )
	{
	  fprintf(stderr, "] KheDoProfileGroupingMonitors returning: uncoordinate ");
	  KheTimeSetDebug(ts, 2, 0, stderr);
	}
	return;
      }
      i = j - 1;
    }
  }

  ** get min_limit, max_limit, history_before, and history_after **
  min_limit = KheLimitActiveIntervalsMonitorMinimum(laim1);
  max_limit = KheLimitActiveIntervalsMonitorMaximum(laim1);
  history_before = KheLimitActiveIntervalsMonitorHistoryBefore(laim1) > 0;
  history_after  = KheLimitActiveIntervalsMonitorHistoryAfter(laim1)  > 0;
  if( laim2 != NULL )
  {
    if( KheLimitActiveIntervalsMonitorMinimum(laim2) > min_limit )
      min_limit = KheLimitActiveIntervalsMonitorMinimum(laim2);
    if( KheLimitActiveIntervalsMonitorMaximum(laim2) < max_limit )
      max_limit = KheLimitActiveIntervalsMonitorMaximum(laim2);
    if( KheLimitActiveIntervalsMonitorHistoryBefore(laim2) > 0 )
      history_before = true;
    if( KheLimitActiveIntervalsMonitorHistoryAfter(laim2)  > 0 )
      history_after = true;
  }

  ** extend the leftmost grouped sequence, if present and too short **
  ts = HaArrayFirst(gfs->time_sets_by_monitor);
  if( ts != NULL && !history_before && KheTimeSetTimeCount(ts) < min_limit )
  {
    for( i = KheTimeSetTimeCount(ts);  i < min_limit && i < tg_count;  i++ )
    {
      index = i;
      if( HaArray(gfs->time_sets_by_monitor, index) != NULL )
	break;  ** we could merge here, but we don't **
      tg = KheLimitActiveIntervalsMonitorTimeGroup(laim1, index, &po);
      t = KheTimeGroupTime(tg, 0);
      HaArrayPut(gfs->time_sets_by_monitor, index, ts);
      KheTimeSetAddTime(ts, t);
    }
    if( i > KheTimeSetTimeCount(ts) )
      KheGroupByFrameGroupTasks(gfs, ts);
  }

  ** extend the rightmost grouped sequence, if present and too short **
  ts = HaArrayLast(gfs->time_sets_by_monitor);
  if( ts != NULL && !history_after && KheTimeSetTimeCount(ts) < min_limit )
  {
    for( i = KheTimeSetTimeCount(ts);  i < min_limit && i < tg_count;  i++ )
    {
      index = (tg_count - 1) - i;
      if( HaArray(gfs->time_sets_by_monitor, index) != NULL )
	break;  ** we could merge here, but we don't **
      tg = KheLimitActiveIntervalsMonitorTimeGroup(laim1, index, &po);
      t = KheTimeGroupTime(tg, 0);
      HaArrayPut(gfs->time_sets_by_monitor, index, ts);
      KheTimeSetAddTime(ts, t);
    }
    if( i > KheTimeSetTimeCount(ts) )
      KheGroupByFrameGroupTasks(gfs, ts);
  }

  ** find all gap intervals **
  prev_ts = NULL;
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim1);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim1, i, &po);
    ts = HaArray(gfs->time_sets_by_monitor, i);
    if( ts == NULL )
    {
      ** find gap i .. j - 1 **
      foll_ts = NULL;
      for( j=i+1; j < KheLimitActiveIntervalsMonitorTimeGroupCount(laim1); j++ )
      {
	foll_ts = HaArray(gfs->time_sets_by_monitor, j);
	if( foll_ts != NULL )
	  break;
      }

      ** merge within the gap (if prev and foll not extensible); carry on **
      if( DEBUG9 )
	fprintf(stderr, "  found gap %d .. %d\n", i, j - 1);
      if( (prev_ts == NULL || KheTimeSetTimeCount(prev_ts) >= max_limit) &&
          (foll_ts == NULL || KheTimeSetTimeCount(foll_ts) >= max_limit) )
	KheMergeOneGap(gfs, laim1, min_limit, max_limit, i, j);
      i = j - 1;
    }
    else
      prev_ts = ts;
  }

  ** try task merging based on the number of tasks at each time **
  KhePlateauMerge(gfs, laim1, min_limit, max_limit);

  if( DEBUG9 )
    fprintf(stderr, "] KheDoProfileGroupingMonitors returning\n");
}
*** */


/* ***
static void KheDoProfileGrouping(KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  KHE_MONITOR m;  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, laimi, laimj;
  int ti, i, j, time_count;  KHE_TIME_SET ts;  KHE_INSTANCE ins;

  ** sort and merge the time sets from Phase 1 **
  ** sti ll to do - in GPost they are all disjoint anyway **
  ** when we do this, we can combine it with the next step **

  ** set up time_sets_by_time, mapping times to time sets where present **
  HaArrayClear(gfs->time_sets_by_time);
  ins = KheSolnIns tance(gfs->soln);
  time_count = KheInstanceTimeCount(ins);
  HaArrayFill(gfs->time_sets_by_time, time_count, NULL);
  HaArrayForEach(gfs->time_sets, ts, i)
  {
    ** HaArrayForEach(ts->times, t, j) **
    for( j = 0;  j < KheTimeSetTimeCount(ts);  j++ )
    {
      ti = KheTimeSetTimeIndex(ts, j);
      HaArrayPut(gfs->time_sets_by_time, ti, ts);
    }
  }
  if( DEBUG4 )
  {
    fprintf(stderr, "  [ time_sets_by_time:\n");
    HaArrayForEach(gfs->time_sets_by_time, ts, i)
    {
      fprintf(stderr, "    %d (%s): ", i, KheTimeId(KheInstanceTime(ins, i)));
      if( ts != NULL )
	KheTimeSetDebug(ts, 2, 0, stderr);
      else
	fprintf(stderr, "NULL\n");
    }
    fprintf(stderr, "  ]\n");
  }

  ** find the suitable limit active intervals monitors **
  HaArrayClear(gfs->suitable_monitors);
  for( i = 0;  i < KheSolnResourceMonitorCount(gfs->soln, gfs->resource);  i++ )
  {
    m = KheSolnResourceMonitor(gfs->soln, gfs->resource, i);
    if( KheMonitorTag(m) == KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG )
    {
      laim = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
      if( KheLimitActiveIntervalsMonitorSuitsProfileGrouping(laim) )
      {
	if( DEBUG7 )
	{
	  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c;
	  c = KheLimitActiveIntervalsMonitorConstraint(laim);
	  fprintf(stderr, "  found suitable monitor, constraint %s\n",
	    KheConstraintId((KHE_CONSTRAINT) c));
	}
	HaArrayAddLast(gfs->suitable_monitors, laim);
      }
    }
  }

  ** find pairs of suitable monitors and do the job for each distinct pair **
  for( i = 0;  i < HaArrayCount(gfs->suitable_monitors);  i++ )
  {
    laimi = HaArray(gfs->suitable_monitors, i);
    for( j = i + 1;  j < HaArrayCount(gfs->suitable_monitors);  j++ )
    {
      laimj = HaArray(gfs->suitable_monitors, j);
      if( KheLimitActiveIntervalsMonitorsCanGroupTogether(laimi, laimj) )
      {
	KheDoProfileGroupingMonitors(gfs, laimi, laimj);
	HaArrayDeleteAndPlug(gfs->suitable_monitors, j);  ** order matters! **
	HaArrayDeleteAndPlug(gfs->suitable_monitors, i);  ** order matters! **
	i--;
	break;
      }
    }
  }

  ** find single suitable monitors and do the job for them **
  HaArrayForEach(gfs->suitable_monitors, laim, i)
    KheDoProfileGroupingMonitors(gfs, laim, NULL);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheGroupSolverDeleteTask(KHE_RCG_SOLVER gs, KHE_TASK task)          */
/*                                                                           */
/*  Delete task from gs.  This involves deleting it from its root            */
/*  task set, and if that set becomes empty, deleting it as well.            */
/*                                                                           */
/*****************************************************************************/

/* *** not currently used
static bool KheGroupSolverDeleteTask(KHE_RCG_SOLVER gs, KHE_TASK task)
{
  KHE_MEET meet;  KHE_TIME t;  int i;  KHE_TIME_INFO ti;
  KHE_ROOT_TASK_SET rts;
  meet = KheTaskMeet(task);
  if( meet != NULL && KheTaskAsst(task) == NULL )
  {
    t = KheMeetAsstTime(meet);
    if( t != NULL )
    {
      ** search for a set of equivalent tasks and delete task if found **
      ti = HaArray(gs->time_info, KheTimeIndex(t));
      HaArrayForEach(ti->root_task_sets, rts, i)
	if( KheRootTaskSetDeleteTask(rts, task) )
	{
	  ** if rts is now empty, delete it too **
	  if( KheTaskSetTaskCount(rts->tasks) == 0 )
	    KheRootTaskSetDelete(gs, rts);
	  return true;
	}
      HnAbort("KheGroupSolverDeleteTask internal error 1:  task not present");
    }
  }
  return false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupSolverDeleteTaskFromRootTaskSet(KHE_RCG_SOLVER gs,          */
/*    KHE_ROOT_TASK_SET rts, KHE_TASK task)                                  */
/*                                                                           */
/*  Delete task from rts (it must be present), and if that makes rts empty,  */
/*  delete it too.                                                           */
/*                                                                           */
/*****************************************************************************/

/* *** use KheRootTaskSetDeleteTask now
static void KheGroupSolverDeleteTaskFromRootTaskSet(KHE_RCG_SOLVER gs,
  KHE_ROOT_TASK_SET rts, KHE_TASK task)
{
  KheRootTaskSetDeleteTask(rts, task);
  if( KheTaskSetTaskCount(rts->tasks) == 0 )
    KheRootTaskSetDelete(gs, rts);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupTasksCombinatorial(KHE_RCG_SOLVER gs,                       */
/*    KHE_TIME_SET ts)                                                       */
/*                                                                           */
/*  As far as possible, ensure that every task lying anywhere in ts is in a  */
/*  group that covers all of ts at least.                                    */
/*                                                                           */
/*****************************************************************************/

/* *** replaced by KheTimeBasedGrouping
static void KheGroupTasksCombinatorial(KHE_RCG_SOLVER gs,
  KHE_TIME_SET ts)
{
  KHE_ROOT_TASK_SET root_ets, ets;  int i;  bool stopping;
  KHE_TASK root_task, task;
  HaArrayClear(gs->tmp_ets_array);
  while( KheFindMinDomainRootTaskSet(gs, ts, false, ts, NULL, NULL, &root_ets) )
  {
    ** gs->tmp_remaining_ts is the set of times still to be covered **
    KheTimeSetClear(gs->tmp_remaining_ts);
    KheTimeSetCopyElements(gs->tmp_remaining_ts, ts);
    KheTimeSetDifference(gs->tmp_remaining_ts, root_ets->times);

    ** gs->tmp_covered_ts is the set of times currently covered **
    KheTimeSetClear(gs->tmp_covered_ts);
    KheTimeSetCopyElements(gs->tmp_covered_ts, root_ets->times);

    ** find other ets's to group with root_ets **
    while( KheFindMinDomainRootTaskSet(gs, gs->tmp_remaining_ts, false, NULL,
	gs->tmp_covered_ts, root_ets, &ets) )
    {
      HaArrayAddLast(gs->tmp_ets_array, ets);
      KheTimeSetDifference(gs->tmp_remaining_ts, ets->times);
      KheTimeSetUnion(gs->tmp_covered_ts, ets->times);
    }

    ** do as many assignments as the group sizes will allow **
    stopping = false;
    do
    {
      root_task = KheTaskSetTask(root_ets->tasks, 0);
      if( KheTaskSetTaskCount(root_ets->tasks) == 1 )
	stopping = true;
      KheGroupSolverDeleteTask(gs, root_task);
      HaArrayForEach(gs->tmp_ets_array, ets, i)
      {
	task = KheTaskSetTask(ets->tasks, 0);
	if( KheTaskSetTaskCount(ets->tasks) == 1 )
	  stopping = true;
	KheGroupSolverDeleteTask(gs, task);
	if( !KheTaskAssign(task, root_task) )
	  HnAbort("KheGroupTasksCombinatorial internal error 1");
      }
      KheGroupSolverAddTask(gs, root_task);
      if( DEBUG3 )
      {
	fprintf(stderr, "  combinatorial group ");
	KheTaskDebug(root_task, 2, 0, stderr);
      }
    } while( !stopping );
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_GROUP_BY_FRAME_SOLVER - solver for group by frame solving            */
/*                                                                           */
/*  NB start_indexes[si] is the smallest i such that                         */
/*  framed_monitors[i]->start_index >= si, or if there is no such            */
/*  index there is no entry for si in start_indexes.                         */
/*                                                                           */
/*  Consequently, the following code will visit all monitors whose           */
/*  frames start at or after si:                                             */
/*                                                                           */
/*  if( si < HaArrayCount(start_indexes) )                                   */
/*    for( i = start_indexes[si];  i < HaArrayCount(framed_monitors);  i++ ) */
/*      visit(framed_monitors[i];                                            */
/*                                                                           */
/*  and the first time the visit comes upon a monitor whose first index      */
/*  is past the interval we are interested in, we can stop.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
typedef struct khe_group_by_frame_solver_rec {

  ** the solve as a whole **
  HA_ARENA			arena;
  KHE_TASKING			tasking;
  KHE_SOLN			soln;
  bool				resource_invariant;
  KHE_FRAME			common_frame;
  KHE_EVENT_TIMETABLE_MONITOR	etm;
  KHE_RESOURCE			resource;
  bool				soln_changed;
  KHE_TASK_SET			task_set;

  ** phase 0 **
  HA_ARRAY_BOOL			provisionally_linked_to_prev;
  HA_ARRAY_BOOL			linked_to_prev;

  ** phase 1 **
  int				max_days;
  ** ***
  ARRAY_KHE_FRAMED_MONITOR	framed_monitors;
  HA_ARRAY_INT			start_indexes;
  *** **
  KHE_GROUP_MONITOR		group_monitor;
  KHE_TIME_SET			successful_times;
  int				last_index;
  int				success_count;
  KHE_TIME_SET			success_ts;
  KHE_TIME_SET			curr_ts;

  ** phase 2 **
  ARRAY_KHE_TIME_SET		time_sets;
  ARRAY_KHE_TIME_SET		time_sets_by_time;
  ARRAY_KHE_TIME_SET		time_sets_by_monitor;
  ARRAY_KHE_LIMIT_ACTIVE_INTERVALS_MONITOR suitable_monitors;
  HA_ARRAY_INT			success_seq;
  HA_ARRAY_INT			curr_seq;
  int				node_count;
  ARRAY_KHE_TASK_SET		task_sets;
  HA_ARRAY_INT			counts;
  ARRAY_KHE_DOMAIN_TASK_SET	domain_task_sets;

} *KHE_GROUP_BY_FRAME_SOLVER;
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheRootTaskSetDeleteTask(KHE_ROOT_TASK_SET rts, KHE_TASK task)      */
/*                                                                           */
/*  If task is present in rts, delete it and return true.  Otherwise change  */
/*  nothing and return false.                                                */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused, probably obsolete
static bool KheRootTaskSetDeleteTask(KHE_ROOT_TASK_SET rts, KHE_TASK task)
{
  int pos, i;  KHE_TIME_INFO ti;
  if( KheTaskSetContainsTask(rts->tasks, task, &pos) )
  {
    KheTaskSetDeleteTask(rts->tasks, task);
    if( rts->pg_active && rts->pg_time_count < rts->solver->pg_max_limit )
      HaArrayForEach(rts->time_info, ti, i)
	if( ti->pg_active )
	  ti->pg_profile--;
    if( KheTaskSetTaskCount(rts->tasks) == 0 )
      KheRootTaskSetDelete(rts);
    return true;
  }
  else
    return false;
}
*** */
